#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/3/7
import datetime

import pytz

from api.annotation.annotation import TaskHandler
from api.handler.crawl.ISimpleScriptCrawl import ISimpleScriptCrawl
from api.model.FactDatasource import FactDatasource
from schedule import TaskHandlerContext
from schedule.test.suite import DataCrawl
from schedule.test.suite.crawl import AbstractSimpleScriptCrawlTestSuite


@TaskHandler("demoSimpleScriptCrawl")
class DemoSimpleScriptCrawl(ISimpleScriptCrawl):

    """CRAWL_python脚本"""
    def do_crawl(self, params: dict):

        date = datetime.datetime.now(pytz.timezone('Asia/Shanghai')).strftime('%Y-%m-%d %H:%M:%S')
        result = {
            "id": 1,
            "tv_name": "CCTV-2",
            "epg_name": "Python-CR",
            "rank": 59,
            "date": date
        }
        if params and params.get("name"):
            name = params.get("name")
            result["name"] = name
            self.task_logger.log(f"CRAWL_PYTHON日志测试，接收到环境变量name:{name}")

        self.task_logger.log("CRAWL_PYTHON日志测试，这里输出普通日志")
        try:
            raise Exception('CRAWL_PYTHON日志测试，这里输出错误日志')
        except Exception as e:
            self.task_logger.error(e)

        # 缓存
        self.set_cache_value('name', 'sobey')
        cache_value = self.get_cache_value('name')
        self.task_logger.log(f"缓存值{cache_value}")

        from api.model import OutputWrapper
        return [OutputWrapper.UPSERT(result)]

    def kill(self):
        self.task_logger.log("CRAWL_PYTHON脚本被终止")


class TestAbstractSimpleScriptCETestSuite(AbstractSimpleScriptCrawlTestSuite):

    def mock_target(self):
        return TaskHandlerContext.load_task_handler("demoSimpleScriptCrawl")

    def mock_fact_datasource(self) -> dict:
        fd = {
            "code": "test_source",
            "connection": "localhost:9200",
            "credentials": None,
            "model": None,
            "name": "test_source",
            "privilege": None,
            "projectCode": "TEST",
            "site": "S1",
            "tags": [
                "masterData"
            ],
            "target": "es_test",
            "ttl": 0,
            "type": "ES"
        }

        fds = {
            'test_source': FactDatasource(**fd)
        }
        return fds

    def mock_crawl(self) -> DataCrawl:
        dc = {
            'site': 'S1',
            'projectCode': 'TEST',
            'code': 'test_script_crawl',
            'outputFdCodes': ['test_source'],
            'params': {
                'name': 'sobey'
            }
        }
        return DataCrawl(**dc)



from api.model.FdInputWrapper import FdInputWrapper


class FdInputPipe:
    """
    输入包装
    """
    __source_fd_codes = None

    def __init__(self, source_fd_codes):
        self.__source_fd_codes = source_fd_codes

    def list_source_fd_codes(self):
        """
        列出所有与之关联的来源FD的code
        :return:
        """
        return self.__source_fd_codes

    def get_fd(self, fd_code):
        """
        返回一个Fd的输入代理对象
        :param fd_code:
        :return: FdInputWrapper对象
        """
        return FdInputWrapper(fd_code)



"""
执行器的接口
"""
from abc import abstractmethod

from api.handler import HandlerEnum
from api.handler.IHandler import IHandler
from api.handler.IKillable import IKillable
from api.handler.ILogAbleHandler import ILogAbleHandler
from api.handler.IProgressAble import IProgressAble


class ITaskHandler(ILogAbleHandler, IHandler, IKillable, IProgressAble):

    @abstractmethod
    def execute(self, params):
        """
        执行任务
        :return:
        """

    @abstractmethod
    def get_execution_message_cache(self):
        """
        获取执行消息的写入器
        :return:
        """

    def write_execution_message(self, execution_message):
        """
        写入执行消息
        :param execution_message:
        :return:
        """
        if self.get_execution_message_cache() is not None:
            self.get_execution_message_cache().add(execution_message)

    def handler_enum(self) -> HandlerEnum:
        return HandlerEnum.NORMAL



import logging
import sys
# 开始使用异步服务器启动 TODO 发现异步在message消息的情况下会异常退出,先不使用
# import gevent.monkey
# gevent.monkey.patch_all()

from flask import Flask
from flask_cors import *
from munch import Munch

import config
from config import annotation

log = logging.getLogger('Ficus')

log.info("=======服务启动开始======")
app = Flask(__name__)
# 解决flask中文乱码
app.config['JSON_AS_ASCII'] = False
# 设置跨域
CORS(app, supports_credentials=True)

# 读取本地配置文件
from config.BootstrapPropertyLoader import load_properties_after_started, init_from_yaml_property, \
    init_from_environ_property, init_profile_from_environ_property

# 先从环境变量中获取profile
init_profile_from_environ_property()

init_from_yaml_property(sys.argv[0])
# 尝试从环境变量中获取 bootstrap里面的信息
init_from_environ_property()

log.info("服务启动,完成本地yaml配置文件读取")

# 注册信息到eureka中
from py_eureka_client import eureka_client

eureka_client.init(eureka_server=config.eureka_default_zone,
                   app_name=config.application_name,
                   # 当前组件的主机名，可选参数，如果不填写会自动计算一个，如果服务和 eureka 服务器部署在同一台机器，请必须填写，否则会计算出 127.0.0.1
                   instance_host=config.server_ip or config.find_host_ip(),
                   instance_port=config.server_port or 5000,
                   # 调用其他服务时的高可用策略，可选，默认为随机
                   ha_strategy=eureka_client.HA_STRATEGY_STICK,
                   status_page_url="/actuator/info",
                   health_check_url="/actuator/health",
                   renewal_interval_in_secs=4,
                   duration_in_secs=12)

log.info("服务启动,完成注册中心注册及心跳")

load_properties_after_started(sys.argv[0])

log.info("服务启动,完成配置中心配置文件读取")

# 这一行不能去掉,目的是引入flask的endpoints,并且位置需要在 app = Flask(__name__) 后面
# 引入views
from remote import remote

app.register_blueprint(remote, url_prefix='/')

log.info("服务启动,完成flask框架启动")

# 先加载registry.yml
import registry

registry.load_registry_properties(sys.argv[0], "registry.yml")

# 加载规则引擎的配置
from rule_engine import RuleEngineInitialize

RuleEngineInitialize.initialize_from_remote()
log.info("服务启动,完成规则引擎的初始化")

# 加载Celery的框架东西
# 获取远程配置中的配置项.判断是否有规则引擎相关的配置
try:
    if annotation.get_value("${celery.enable:false}"):
        import cloudcelery

        log.info("服务启动,完成celery的预启动")
except:
    # 无所谓
    pass

# 预先加载根目录下的这个模块,这样才能在程序启动后,自动注册执行器
try:
    import handlers
except Exception as e:
    log.error(f"加载执行器实例错误:{str(e)}")
    exit(-1)  # 程序直接退出
log.info("服务启动,完成处理器的预加载")

# 程序启动后,判断是否需要注册执行器
from registry.LoadOnRegistryLoader import registry_after_started

registry_after_started()

log.info("服务启动,完成处理器的注册")

from schedule.utils.log.TaskLogCleanScheduler import TaskLogCleanScheduler

log_cleaner = TaskLogCleanScheduler()
log_cleaner.start()
log.info("服务启动,完成日期清理线程的启动")

log.info(f"服务启动:{config.server_ip or config.find_host_ip()}:{config.server_port or 5000}")
log.info("======服务启动完毕======")

from multiprocessing import Process


class CeleryProcess(Process):  # 继承Process类
    def __init__(self, name):
        super(CeleryProcess, self).__init__()
        self.name = name

    def run(self):
        import uuid
        argv = [
            'worker',
            '--loglevel=INFO',
            f'--hostname={config.actor_name}@{uuid.uuid4()}'
        ]
        import cloudcelery
        cloudcelery.celery.worker_main(argv)


def run_flask():
    """
    单独启动flask
    :return:
    """
    from config import server_port
    from gevent import pywsgi
    # import gevent.monkey
    # gevent.monkey.patch_all()
    server = pywsgi.WSGIServer(('0.0.0.0', server_port), app)
    server.serve_forever()


def run_flask_with_celery():
    """
    启动flask和celery
    :return:
    """
    p2 = CeleryProcess('Celery')
    p2.start()

    run_flask()
    p2.join()


def run_with_auto_select():
    if annotation.get_value("${celery.enable:false}"):
        run_flask_with_celery()
    else:
        run_flask()



import logging

from flask import request, jsonify
from munch import Munch

from . import remote
from schedule import TriggerActor

log = logging.getLogger('Ficus')

@remote.route('/remote/ta-service/<int:jobId>', methods=['DELETE'])
def stop_job(jobId):
    return TriggerActor.stop_job(jobId).to_json()

@remote.route('/remote/ta-service/<int:jobId>/<int:logId>', methods=['DELETE'])
def stop_task(jobId,logId):
    is_executing = request.args.get("isExecuting",default="false")
    return TriggerActor.stop_task(jobId,logId,'true'==is_executing).to_json()

@remote.route('/remote/ta-service', methods=['POST'])
def handle_trigger():
    taskParam = request.json
    body = Munch(taskParam)
    log.debug(f"触发了一次请求:{taskParam}")
    """
    {
  "actorBlockStrategy": "SERIAL_EXECUTION",
  "actorHandler": "multipleArticleMessageCEHandler",
  "actorParams": {
    "code_": "locus.multiple.article.message.ce-0",
    "projectCode_": "LOCUS",
    "site_": "S1"
  },
  "jobId": 20081,
  "jobType": "BEAN",
  "limitTimes": -1,
  "logId": 55,
  "scriptJobRemark": "None",
  "scriptJobSource": "None",
  "updateTime": "2018-02-04 10:35:39"
}
    """

    return TriggerActor.handle_trigger(body).to_json()


@remote.route('/remote/ta-service/ping', methods=['GET'])
def ping():
    return jsonify("pong")

@remote.route('/remote/ta-service/idle/<int:job_id>', methods=['GET'])
def idle(job_id:int):
    return jsonify(TriggerActor.idle(job_id))


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/3/7
import datetime

import pytz

from api.annotation.annotation import TaskHandler
from api.handler.crawl.AbstractBatchCrawl import AbstractBatchCrawl
from api.model.BatchOutputPipe import BatchOutputPipe
from api.model.FactDatasource import FactDatasource
from config.annotation import Value
from schedule import TaskHandlerContext
from schedule.test.suite import DataCrawl
from schedule.test.suite.crawl import AbstractBatchCrawlTestSuite


@TaskHandler("demoAbstractBatchCrawl")
class DemoAbstractBatchCrawl(AbstractBatchCrawl):

    @Value("${server.port:unknown}")
    def server_port(self):
        pass

    """CRAWL_python脚本"""
    def do_crawl(self, output_stream: BatchOutputPipe, params: dict):
        print(self.server_port())

        date = datetime.datetime.now(pytz.timezone('Asia/Shanghai')).strftime('%Y-%m-%d %H:%M:%S')
        result = {
            "id": 3,
            "tv_name": "CCTV-2",
            "epg_name": "Python-CR",
            "rank": 59,
            "date": date
        }
        if params and params.get("name"):
            name = params.get("name")
            result["name"] = name
            self.task_logger.log(f"CRAWL_PYTHON日志测试，接收到环境变量name:{name}")

        self.task_logger.log("CRAWL_PYTHON日志测试，这里输出普通日志")
        try:
            raise Exception('CRAWL_PYTHON日志测试，这里输出错误日志')
        except Exception as e:
            self.task_logger.error(e)

        # 缓存
        self.set_cache_value('name', 'sobey')
        cache_value = self.get_cache_value('name')
        self.task_logger.log(f"缓存值:{cache_value}")

        output_stream.output_for_upsert(result)

    def kill(self):
        self.task_logger.log("CRAWL_PYTHON脚本被终止")


class TestAbstractBatchCrawlTestSuite(AbstractBatchCrawlTestSuite):

    def mock_target(self):
        return TaskHandlerContext.load_task_handler("demoAbstractBatchCrawl")

    def mock_fact_datasource(self) -> dict:
        fd = {
            "code": "test_source",
            "connection": "localhost:9200",
            "credentials": None,
            "model": None,
            "name": "test_source",
            "privilege": None,
            "projectCode": "TEST",
            "site": "S1",
            "tags": [
                "masterData"
            ],
            "target": "es_test",
            "ttl": 0,
            "type": "ES"
        }

        fds = {
            'test_source': FactDatasource(**fd)
        }
        return fds

    def mock_crawl(self) -> DataCrawl:
        dc = {
            'site': 'S1',
            'projectCode': 'TEST',
            'code': 'test_script_crawl',
            'outputFdCodes': ['test_source'],
            'params': {
                'name': 'sobey'
            }
        }
        return DataCrawl(**dc)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
import logging

from factdatasource.execptions import NotSupportedFDException
from factdatasource.persistence.AbstractFactDatasourceContext import AbstractFactDatasourceContext

log = logging.getLogger('Ficus')


class FileFactDatasourceContext(AbstractFactDatasourceContext):

    def size(self):
        """
        返回数据总长度
        :return: 数据条数:long
        """
        raise NotSupportedFDException('File不支持该操作')

    def is_empty(self):
        """
        返回是否存在数据
        :return: boolean
        """
        raise NotSupportedFDException('File不支持该操作')

    def collect(self, size: int):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :return: list
        """
        raise NotSupportedFDException('File不支持该操作')

    def collect_conditions(self, size: int, condition_groups: list):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :param condition_groups: 查询条件
        :return: list
        """
        raise NotSupportedFDException('File不支持该操作')

    def query(self, query: str, parameters: dict):
        """
        使用查询语句查询数据
        :param query: 查询语句
        :param parameters: 查询参数
        :return: Page
        """
        raise NotSupportedFDException('File不支持该操作')

    def _single_thread_inserts(self, table: str, result_list: list):
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param result_list: 要保存的数据
        :return:
        """
        raise NotSupportedFDException('File不支持该操作')

    def _single_thread_updates(self, table: str, result_list: list):
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要修改的数据
        :return:
        """
        raise NotSupportedFDException('File不支持该操作')

    def _single_thread_inserts_or_updates(self, table: str, result_list: list):
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        raise NotSupportedFDException('File不支持该操作')

    def delete_all(self):
        """
        清空数据
        :return:
        """
        raise NotSupportedFDException('File不支持该操作')

    def delete(self, query: str):
        """
        根据删除语句删除数据,query是完整的删除语句
        :return:
        """
        raise NotSupportedFDException('File不支持该操作')

    def delete_conditions(self, condition_groups: list):
        raise NotSupportedFDException('File不支持该操作')



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/3/11


class SimpleNode(object):
    def __int__(self, id: str=None, group: str=None, type: str=None, resourceId: str=None, tags: str=None):
        self.__id = id
        self.__group = group
        self.__type = type
        self.__resourceId = resourceId
        self.__tags = tags

    @property
    def id(self):
        return self.__id

    @id.setter
    def id(self, value: str):
        self.__id = value

    @property
    def group(self):
        return self.__group

    @group.setter
    def group(self, value: str):
        self.__group = value

    @property
    def type(self):
        return self.__type

    @type.setter
    def type(self, value: str):
        self.__type = value

    @property
    def resourceId(self):
        return self.__resourceId

    @resourceId.setter
    def resourceId(self, value: str):
        self.__resourceId = value

    @property
    def tags(self):
        return self.__tags

    @tags.setter
    def tags(self, value: str):
        self.__tags = value


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/2/21
from redis import StrictRedis

from api.model.FactDatasource import FactDatasourceTypeEnum
from factdatasource.dao.FactDatasource import Datasource
from factdatasource.dao.MultipleDatasourceHolder import get_multiple_datesource
from libs.utils import Singleton


class RedisDatasourceDao(Singleton):

    @property
    def datasource(self) -> Datasource:
        data_source = get_multiple_datesource(FactDatasourceTypeEnum.REDIS)
        return data_source

    @property
    def client(self) -> StrictRedis:
        return self.datasource.get_client()




import json
import logging
from threading import Thread

from flask import request
from munch import Munch

from api.exceptions import ServiceInnerException
from api.handler.IAsyncAble import IAsyncAble
from api.handler.script.ScriptPythonTaskHandler import ScriptPythonTaskHandler
from api.model.ResultVO import FAIL, FAIL_CODE, ResultVO
from api.model.AsyncServiceRequest import AcceptType
from api.model.AsyncServiceResponse import AsyncServiceResponse, FAILED_CODE
from client import ScheduleJobTaskLogClient, JobScheduleClient
from schedule import TaskThreadHolder, TaskHandlerContext
from schedule.TaskThread import TaskThread
from . import remote

log = logging.getLogger('Ficus')


def build_task_param(schedule_job_info, job_task_log, task_process_log_id_, ficus_param):
    """
    构造taskParam
    :param schedule_job_info:
    :param job_task_log:
    :param task_process_log_id_:
    :param ficus_param:
    :return:
    """
    global shardingArr
    shardingParam: str = job_task_log.shardingParam
    if shardingParam is not None:
        shardingArr = shardingParam.split("/")

    m = {
        "actorBlockStrategy": schedule_job_info.actorBlockStrategy,
        "actorName": schedule_job_info.actorName,
        "actorHandler": schedule_job_info.actorHandler,
        "actorParams": ficus_param,
        "jobId": schedule_job_info.id,
        "jobType": schedule_job_info.jobType,
        "limitTimes": schedule_job_info.limitTimes,
        "logId": job_task_log.id,
        "scriptJobRemark": schedule_job_info.scriptJobRemark,
        "scriptJobSource": schedule_job_info.scriptJobSource,
        "updateTime": "2018-02-04 10:35:39",
        "triggerTime": job_task_log.triggerTime,
        "retryLogId": job_task_log.retryLogId,
        "processLogId": task_process_log_id_,
        "retryTimes": job_task_log.retryTimes if "retryTimes" in job_task_log else None,
        "shardIndex": -1 if shardingParam is None else shardingArr[0],
        "shardTotal": -1 if shardingParam is None else shardingArr[1]
    }
    return Munch(m)


def async_do_finish(task_log_id, ficus_param: dict, headers: dict, async_service_response: AsyncServiceResponse,
                    task_thread: TaskThread, task_handler: IAsyncAble,
                    task_param: Munch):
    """
    异步处理结果
    :param task_log_id:
    :param ficus_param:
    :param headers:
    :param async_service_response:
    :param task_thread:
    :param task_handler:
    :param task_param:
    :return:
    """
    if async_service_response.task_state == FAILED_CODE:
        headers["__error__"] = async_service_response.task_desc

    try:
        executeResult = task_handler.do_finish(task_log_id, async_service_response.response_data, headers, ficus_param,
                                               async_service_response.task_state)
        if executeResult is None:
            executeResult = FAIL
    except Exception as e:
        executeResult = ResultVO(FAIL_CODE, str(e))

    try:
        task_thread.handle_execute_result(task_param, executeResult)
    except Exception as e:
        stopResult = ResultVO(FAIL_CODE, f" 处理logId:{task_log_id} 出现严重错误:{str(e)}")
        ScheduleJobTaskLogClient.update_task_status_to_finished(task_log_id, stopResult.to_dict(), True)


@remote.route('/async/response/<int:job_id>/<int:task_log_id>', methods=['POST'])
def do_finish(job_id: int, task_log_id: int, accept_type: AcceptType = AcceptType.APPLICATION_JSON):
    """
    用户接收消息的回调,不再ficus-core中写,而是在这写
    是因为避免返回的数据太大,不适合从ficus-core中再过一次
    :param job_id:
    :param task_log_id:
    :param accept_type:
    :return:
    """

    log.info(f"接收到任务回调,jobId:{job_id} logId:{task_log_id}")

    job_task_log = ScheduleJobTaskLogClient.get_task_log_by_id(task_log_id)

    if job_task_log is None:
        raise ServiceInnerException(f"任务:{task_log_id}没有找到数据,无法回调")

    if 1 != job_task_log.executeStatus:
        raise ServiceInnerException(f"任务:{task_log_id}执行状态为{job_task_log.executeStatus},无法回调")

    ficus_param = json.loads(job_task_log.actorParam)

    ficus_param["__jobId__"] = job_id
    ficus_param["__logId__"] = task_log_id
    ficus_param["__triggerTime__"] = job_task_log.triggerTime

    taskProcessLogId_ = ficus_param["taskProcessLogId_"] if "taskProcessLogId_" in ficus_param else None

    if taskProcessLogId_ is not None:
        ficus_param["__processLogId__"] = str(taskProcessLogId_)

    schedule_job_info = JobScheduleClient.find_by_site_and_code(ficus_param["site_"], ficus_param["projectCode_"],
                                                                ficus_param["code_"])

    task_param = build_task_param(schedule_job_info, job_task_log, taskProcessLogId_, ficus_param)

    task_thread: TaskThread = TaskThreadHolder.load_task_thread(job_id)

    task_handler = task_thread.get_handler() if task_thread is not None else None

    if task_handler is None:
        # 需要重新实例化
        task_handler = TaskHandlerContext.load_task_handler(task_param.actorHandler)

    # 有可能是脚本类型的异步任务，尝试加载一下
    if task_handler is None:
        task_handler = load_async_script_task_handler()

    if task_thread is None:
        task_thread = TaskThreadHolder.registry_task_thread(job_id, task_handler, "异步任务构造taskThread")

    if isinstance(task_handler, ScriptPythonTaskHandler):
        task_handler = task_handler.current_task

    if not isinstance(task_handler, IAsyncAble):
        # 说明不是异步的任务,抛错
        raise ServiceInnerException(f"任务:{task_log_id}执行类型不是异步执行器,无法回调")

    data = request.get_data()
    response_data = None
    if data is not None:
        data = data.decode("utf-8");
        if accept_type == AcceptType.APPLICATION_JSON:
            response_data = Munch(json.loads(data))
        else:
            response_data = Munch({'response:': data})
    headers = dict(request.headers)
    async_service_response = task_handler.transform_response(response_data, headers)

    # 异步执行结果回调
    t = Thread(target=async_do_finish, args=(
        task_log_id, ficus_param, headers, async_service_response, task_thread, task_handler, task_param))
    t.start()

    return "success"

@remote.route('/async/response/<int:job_id>/<int:task_log_id>/plain', methods=['POST'])
def do_finish_from_plain(job_id: int, task_log_id: int):
    return do_finish(job_id, task_log_id, AcceptType.TEXT_PLAIN)

def load_async_script_task_handler():
    from config.annotation import Value
    from api.handler.script import ScriptPythonFactory

    @Value("${actor.name:unknown}")
    def actor_name():
        pass

    # region 过滤函数
    def filter_job(item) -> bool:
        if item.jobType == "PYTHON" and item.scriptJobSource is not None and len(item.scriptJobSource) > 0:
            return True
        return False

    jobs = JobScheduleClient.get_jobs(actor_name())
    if jobs is None or len(jobs) == 0:
        return None

    jobs = list(filter(filter_job, jobs))
    if jobs is None or len(jobs) == 0:
        return None

    # 实例化Handler
    job_info = jobs.pop(0);
    script_task_handler:ScriptPythonTaskHandler = ScriptPythonTaskHandler(job_info.id, job_info.updateTime, job_info.scriptJobSource)
    instance = ScriptPythonFactory.load_instance(f"ScriptPython{job_info.id}", job_info.scriptJobSource, "IHandler")
    script_task_handler.current_task = instance

    return script_task_handler




import logging

from api.exceptions import IllegalArgumentException
from client import ComputeExecutionClient, DataCrawlClient
from config.annotation import Value
from factdatasource import FactDatasourceProxyService

log = logging.getLogger('Ficus')


@Value("${actor.name:unknown}")
def actor_name():
    pass


def check_ficus_online():
    """
    检查ficus服务是否存在
    :return:
    """
    from client import check_instance_avaliable
    from api.exceptions import ServiceNoInstanceException
    try:
        check_instance_avaliable()
        return True
    except ServiceNoInstanceException:
        return False


def createNewComputeExecutions(registry, taskHandlerNames, site, projectcode):
    """
    尝试创建ce
    :param taskHandlerNames:
    :return:
    """
    if "computeexecutions" not in registry:
        return

    # region 先尝试添加ce
    needRegistryList = []

    # 表示有ce需要注册
    index = 1
    for computeexecution in registry["computeexecutions"]:

        # region 补默认值
        if "code" not in computeexecution or computeexecution["code"] == "" or computeexecution["code"] is None:
            computeexecution["code"] = f"{actor_name()}-{index}"

        if "execution" not in computeexecution or computeexecution["execution"] is None:
            computeexecution["execution"] = {"schedule": "0 0 0 1 1 ? 2099", "limits": -1, "params": []}
        else:
            if "schedule" not in computeexecution["execution"] or computeexecution["execution"]["schedule"] is None or \
                    computeexecution["execution"]["schedule"] == "":
                computeexecution["execution"]["schedule"] = "0 0 0 1 1 ? 2099"
            if "limits" not in computeexecution["execution"] or computeexecution["execution"]["limits"] is None:
                computeexecution["execution"]["limits"] = -1
            if "params" not in computeexecution["execution"] or computeexecution["execution"]["params"] is None:
                computeexecution["execution"]["params"] = {}
            if "retryTimes" not in computeexecution["execution"] or computeexecution["execution"]["retryTimes"] is None:
                computeexecution["execution"]["retryTimes"] = 0

        if "managed" not in computeexecution or computeexecution["managed"] is None:
            computeexecution["managed"] = {"type": "STANDBY", "serviceName": "STANDBY", "routing": "RANDOM"}
        else:
            if "type" not in computeexecution["managed"] or computeexecution["managed"]["type"] is None or \
                    computeexecution["managed"]["type"] == "":
                computeexecution["managed"]["type"] = "STANDBY"
            if "serviceName" not in computeexecution["managed"] or computeexecution["managed"]["serviceName"] is None or \
                    computeexecution["managed"]["serviceName"] == "":
                computeexecution["managed"]["serviceName"] = "STANDBY"
            if "routing" not in computeexecution["managed"] or computeexecution["managed"]["routing"] is None or \
                    computeexecution["managed"]["routing"] == "":
                computeexecution["managed"]["routing"] = "RANDOM"

        if "outputs" not in computeexecution or computeexecution["outputs"] is None:
            computeexecution["outputs"] = []
        # endregion

        index = index + 1

        # 开始进行validate的校验

        # region 第一轮的校验主要是做数据的验证
        if "name" not in computeexecution or computeexecution["name"] == "":
            raise IllegalArgumentException(f"自注册{computeexecution['code']}失败,名称为空")

        if "handler" not in computeexecution or computeexecution["handler"] == "":
            raise IllegalArgumentException(f"自注册{computeexecution['code']}失败,执行器为空")

        if "inputs" not in computeexecution or len(computeexecution["inputs"]) == 0:
            raise IllegalArgumentException(f"自注册{computeexecution['code']}失败,输入FD为空")

        if computeexecution["handler"] not in taskHandlerNames.keys():
            raise IllegalArgumentException(f"自注册{computeexecution['code']}失败,{computeexecution['handler']} 不存在")
        # endregion

        # region 第二轮的校验是判断东西在不在
        if ComputeExecutionClient.exists(site, projectcode, computeexecution["code"]):
            # 说明已经存在了, 那么就忽略
            continue
        # endregion

        # region 第三轮的校验是对应的fd,是否存在,因为自注册的handler应该不会太多, 它对应的input和output也不会太多,因此,这里就不再先做一次多handler的聚合了
        fds = list()
        fds = fds + computeexecution["inputs"]
        fds += computeexecution["outputs"]

        if not FactDatasourceProxyService.fd_client_proxy().exists_fds(fds):
            # 只要有一个fd不存在,就返回false
            # 表示校验异常了.
            raise IllegalArgumentException(f"自注册{computeexecution['code']}失败,{fds}不存在")
        # endregion

        # 到这的都是要的的
        needRegistryList.append(computeexecution)

    # 上面都是校验,如果到这了就是符合条件的,需要添加的ce了
    for computeexecution in needRegistryList:
        # 构造层ce对象,然后添加
        if taskHandlerNames[computeexecution["handler"]]:
            computeexecution["execution"]["params"]["__isMessageHandler__"] = "true"

        ce = {"type": "JAVA", "sourceFdCodes": computeexecution["inputs"], "outputFdCodes": computeexecution["outputs"],
              "code": computeexecution["code"], "description": computeexecution["description"],
              "projectCode": projectcode, "site": site, "target": computeexecution["handler"],
              "name": computeexecution["name"], "schedule": computeexecution["execution"]["schedule"],
              "limitTimes": computeexecution["execution"]["limits"],
              "params": computeexecution["execution"]["params"],
              "retryTimes": computeexecution["execution"]["retryTimes"],
              "executionType": computeexecution["managed"]["type"],
              "executionServiceName": computeexecution["managed"]["serviceName"],
              "routing": computeexecution["managed"]["routing"]}
        ComputeExecutionClient.add_compute_execution(ce)
        log.info(f"自动注册:{computeexecution['code']} 的ce")
    # endregion


def createNewCrawls(registry, taskHandlerNames, site, projectcode):
    """
    尝试创建crawl
    :param taskHandlerNames:
    :return:
    """
    if "crawls" not in registry:
        return

    # 表示有crawl需要注册
    # region 尝试添加crawl
    needRegistryList = []

    # 表示有crawl需要注册
    index = 1
    for crawl in registry["crawls"]:

        # region 补默认值
        if "code" not in crawl or crawl["code"] == "":
            crawl["code"] = f"{actor_name()}-{index}"

        if "execution" not in crawl:
            crawl["execution"] = {"schedule": "0 0 0 1 1 ? 2099", "limits": -1, "params": []}
        else:
            if "schedule" not in crawl["execution"]:
                crawl["execution"]["schedule"] = "0 0 0 1 1 ? 2099"
            if "limits" not in crawl["execution"]:
                crawl["execution"]["limits"] = -1
            if "retryTimes" not in crawl["execution"] or crawl["execution"][
                "retryTimes"] is None:
                crawl["execution"]["retryTimes"] = 0
            if "params" not in crawl["execution"]:
                crawl["execution"]["params"] = {}

        if "managed" not in crawl:
            crawl["managed"] = {"type": "STANDBY", "serviceName": "STANDBY", "routing": "RANDOM"}
        else:
            if "type" not in crawl["managed"]:
                crawl["managed"]["type"] = "STANDBY"
            if "serviceName" not in crawl["managed"]:
                crawl["managed"]["serviceName"] = "STANDBY"
            if "routing" not in crawl["managed"]:
                crawl["managed"]["routing"] = "RANDOM"

        if "outputs" not in crawl:
            crawl["outputs"] = []
        # endregion

        index = index + 1

        # 开始进行validate的校验

        # region 第一轮的校验主要是做数据的验证
        if "name" not in crawl or crawl["name"] == "":
            raise IllegalArgumentException(f"自注册{crawl['code']}失败,名称为空")

        if "handler" not in crawl or crawl["handler"] == "":
            raise IllegalArgumentException(f"自注册{crawl['code']}失败,执行器为空")

        if "outputs" not in crawl or len(crawl["outputs"]) == 0:
            raise IllegalArgumentException(f"自注册{crawl['code']}失败,输出FD为空")

        if crawl["handler"] not in taskHandlerNames.keys():
            raise IllegalArgumentException(f"自注册{crawl['code']}失败,{crawl['handler']} 不存在")
        # endregion

        # region 第二轮的校验是判断东西在不在
        if DataCrawlClient.exists(site, projectcode, crawl["code"]):
            # 说明已经存在了, 那么就忽略
            continue
        # endregion

        # region 第三轮的校验是对应的fd,是否存在,因为自注册的handler应该不会太多, 它对应的input和output也不会太多,因此,这里就不再先做一次多handler的聚合了
        fds = list()
        fds += crawl["outputs"]

        if not FactDatasourceProxyService.fd_client_proxy().exists_fds(fds):
            # 只要有一个fd不存在,就返回false
            # 表示校验异常了.
            raise IllegalArgumentException(f"自注册{crawl['code']}失败,{fds}不存在")
        # endregion

        # 到这的都是要的的
        needRegistryList.append(crawl)

    # 上面都是校验,如果到这了就是符合条件的,需要添加的ce了
    for crawl in needRegistryList:
        # 构造层ce对象,然后添加
        if taskHandlerNames[crawl["handler"]] :
            crawl["execution"]["params"]["__isMessageHandler__"] = "true"

        cr = {"type": "CUSTOM",
              "outputFdCodes": crawl["outputs"],
              "code": crawl["code"], "description": crawl["description"],
              "projectCode": projectcode, "site": site, "target": crawl["handler"],
              "name": crawl["name"], "schedule": crawl["execution"]["schedule"],
              "limitTimes": crawl["execution"]["limits"],
              "retryTimes": crawl["execution"]["retryTimes"],
              "params": crawl["execution"]["params"],
              "executionType": crawl["managed"]["type"],
              "executionServiceName": crawl["managed"]["serviceName"],
              "routing": crawl["managed"]["routing"]}
        DataCrawlClient.add_crawl(cr)
        log.info(f"自动注册:{crawl['code']} 的crawl")
    # endregion


def registry_after_started():
    """
    服务启动后,检测是否需要注册服务
    :return:
    """

    # 说明有yml文件,开始进行处理
    from registry import REGISTRY_PROPERTIES
    registry = REGISTRY_PROPERTIES

    if registry is None or len(registry) == 0:
        # 说明没有yml文件,不进行注册处理
        log.info("没有配置registry.yml,不进行执行器自动注册")
        return

    if "site" not in registry or "dataproject" not in registry:
        log.info("没有配置site或dataproject,不进行执行器自动注册")
        return

    # 说明还是配置了自注册的文件的,开始处理自注册的文件

    if "computeexecutions" not in registry and "crawls" not in registry:
        # 表示两种都没得,那也是直接返回
        log.info("没有需要注册的ce或crawl,忽略自动注册")
        return

    if not check_ficus_online():
        log.warn("启动时,没有可用的ficus服务,忽略自动注册")
        return

    # 这里开始正式的进行处理

    # 找到这个jar里面的所有的 TaskHandler
    from api.annotation.annotation import TASK_HANDLERS
    if len(TASK_HANDLERS) == 0:
        log.info("无可用的执行器,忽略自动注册")
        return

    # 我们假设 默认 的 site和project都是存在的
    # 找到所有的taskHandler的名称,key是名字,value是否是messageHandler
    taskHandlerNames = {}
    from api.handler.message.IMessageHandler import IMessageHandler
    for name, handler in TASK_HANDLERS.items():
        taskHandlerNames[name] = isinstance(handler, IMessageHandler)

    createNewComputeExecutions(registry, taskHandlerNames, registry["site"], registry["dataproject"])

    createNewCrawls(registry, taskHandlerNames, registry["site"], registry["dataproject"])



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/2/25
import random
import threading
import time
from unittest import TestCase

from factdatasource.dao.FactDatasource import customer_dao_context_holder as holder


def run_1(num: str):
    """
    不同线程操作不用的num
    :param num:
    :return:
    """
    holder.set_source(num)
    print(f'{threading.current_thread().name}:set_{num}->{num}:{holder.get_source()}')
    print(f'{threading.current_thread().name}:set_{num}->True:{holder.is_using(num)}')
    holder.clear_source()
    print(f'{threading.current_thread().name}:clear_{num}->None:{holder.get_source()}')
    print(f'{threading.current_thread().name}:clear_{num}->False:{holder.is_using(num)}')


def run_2(num: str):
    """
    不同线程操作相同num
    :param num:
    :return:
    """
    time.sleep(random.random())
    holder.set_source(num)
    print(f'{threading.current_thread().name}:set_{num}->{num}:{holder.get_source()}')
    print(f'{threading.current_thread().name}:set_{num}->{holder.is_using(num)}->{holder.test()}')
    time.sleep(random.random())

    holder.clear_source()
    print(f'{threading.current_thread().name}:clear_{num}->None:{holder.get_source()}')
    print(f'{threading.current_thread().name}:clear_{num}->{holder.is_using(num)}->{holder.test()}')


class TestCustomerDaoContextHolder(TestCase):

    def test_holder_run_1(self):
        for i in range(0, 10):
            threading.Thread(target=run_1, name=f'线程{i}', args=(str(i),)).start()

    def test_holder_run_2(self):
        for i in range(0, 20):
            threading.Thread(target=run_2, name=f'线程{i}', args=('test',)).start()







from datetime import datetime
from api.model.Condition import Condition, CalculationOperator, ConditionGroup, LogicalOperator
from factdatasource.execptions import FDExecuteException
from munch import Munch


class SqlableDeleteCondition(object):
    """
    能生成删除语句的
    """

    def delete_condition_sql(self, table: str, condition_groups: list) -> str:
        return self.__condition_sql(table, condition_groups, "DELETE")

    def search_condition_sql(self, table: str, condition_groups: list) -> str:
        return self.__condition_sql(table, condition_groups, "SEARCH")

    def __condition_sql(self, table: str, condition_groups: list, method: str) -> str:
        """
        生成条件语句
        这里尽量用的是标准的SQL语法,如果有数据库不一样,就需要重载
        :param table:
        :param condition_groups:
        :param method:
        :return:
        """
        if method == "DELETE":
            sql = f"DELETE FROM `{table}` "
        else:
            sql = f"SELECT * FROM `{table}` "
        if condition_groups is None or len(condition_groups) == 0:
            return sql

        # 说明有查询条件组
        sql = sql + "WHERE "

        for index, condition_group in enumerate(condition_groups):
            if isinstance(condition_group, ConditionGroup):
                if index != 0:
                    sql = sql + f" {str(condition_group.operator.value)} "

                sql = sql + "("
                # 内部条件
                if condition_group.conditions is None or len(condition_group.conditions) == 0:
                    sql = sql + "1=1"
                else:
                    for jndex, condition in enumerate(condition_group.conditions):
                        if jndex != 0:
                            sql = sql + f" {condition.logical_operator.value} "
                        # 加入条件
                        sql = sql + f"`{condition.key}` {condition.calculation_operator.value} "
                        # 加入值
                        sql = sql + self.__generate_simple_value(condition.value)

                sql = sql + ")"
            else:
                if not isinstance(condition_group, dict):
                    raise FDExecuteException("查询条件格式错误")
                else:
                    try:
                        if index != 0:
                            operator = LogicalOperator(condition_group['operator'])
                            sql = sql + f" {str(operator.value)} "

                        sql = sql + "("
                        # 内部条件
                        if condition_group['conditions'] is None or len(condition_group['conditions']) == 0:
                            sql = sql + "1=1"
                        else:
                            for jndex, condition in enumerate(condition_group['conditions']):
                                if jndex != 0:
                                    if LogicalOperator.OR.name == condition['logicalOperator']:
                                        sql = sql + f" {LogicalOperator.OR.value} "
                                    else:
                                        sql = sql + f" {LogicalOperator.AND.value} "
                                # 加入条件
                                if CalculationOperator.EQUAL.name == condition['calculationOperator']:
                                    sql = sql + f"`{condition['key']}` {CalculationOperator.EQUAL.value} "
                                elif CalculationOperator.NOTEQUAL.name == condition['calculationOperator']:
                                    sql = sql + f"`{condition['key']}` {CalculationOperator.NOTEQUAL.value} "
                                elif CalculationOperator.GT.name == condition['calculationOperator']:
                                    sql = sql + f"`{condition['key']}` {CalculationOperator.GT.value} "
                                elif CalculationOperator.GTE.name == condition['calculationOperator']:
                                    sql = sql + f"`{condition['key']}` {CalculationOperator.GTE.value} "
                                elif CalculationOperator.LT.name == condition['calculationOperator']:
                                    sql = sql + f"`{condition['key']}` {CalculationOperator.LT.value} "
                                elif CalculationOperator.LTE.name == condition['calculationOperator']:
                                    sql = sql + f"`{condition['key']}` {CalculationOperator.LTE.value} "
                                elif CalculationOperator.IN.name == condition['calculationOperator']:
                                    sql = sql + f"`{condition['key']}` {CalculationOperator.IN.value} "
                                elif CalculationOperator.LIKE.name == condition['calculationOperator']:
                                    sql = sql + f"`{condition['key']}` {CalculationOperator.LIKE.value} "

                                # 加入值
                                sql = sql + self.__generate_simple_value(condition['value'])

                        sql = sql + ")"
                    except Exception:
                        raise FDExecuteException("查询条件格式错误")
        return sql

    def __generate_simple_value(self, value) -> str:
        """
        构造值的条件
        :param value:
        :return:
        """
        sql = ""

        if value is None:
            sql = "null"
        elif isinstance(value, str):
            sql = sql + f'"{value}"'
        elif isinstance(value, datetime):
            sql = sql + f'"{value.strftime("%Y-%m-%d %H:%M:%S")}"'
        elif isinstance(value, list) or isinstance(value, set):
            sql = sql + "("
            for index, obj in enumerate(value):
                sql = sql + self.__generate_simple_value(obj)
                if index != len(value) - 1:
                    sql = sql + ","
            sql = sql + ")"
        else:
            # 忽略是Map的情况
            sql = sql + f"{value}"
        return sql



from api.handler.IProgressAble import IProgressAble
from api.handler.script import ScriptHandlerHolder


class IProgressAbleScript(IProgressAble):

    def update_task_progress(self, progress: float):
        handler = ScriptHandlerHolder.cacheHandler()
        handler.update_task_progress(progress)



import threading
from abc import abstractmethod
from collections import defaultdict
from datetime import datetime

from api.handler.ICacheAbleHandler import ICacheAbleHandler
from api.handler.ITaskHandler import ITaskHandler
from api.handler.outputer.ISimpleOutputer import ISimpleOutputer
from api.model.ResultVO import *
from client import DataCrawlClient
from schedule.utils.log import TaskLogFileAppender


class AbstractSimpleCrawl(ITaskHandler, ISimpleOutputer, ICacheAbleHandler):
    """
        简单的Crawl,只需要继承这个类,并且实现do_crawl方法即可
    """

    # 由于这个的实现有可能是 单例的,所以要使用ThreadLocal
    def __init__(self):
        super().__init__()
        self.__code_local_host = threading.local()
        self.__process_id_local = threading.local()
        self.__task_log_id_local = threading.local()
        self.__execution_message_local = threading.local()
        self.killed = False

    def execute(self, params: dict):
        """
        简单的方式的Crawl,就是 实现上 直接返回数据集合.然后由这个抽象类来负责往fd上去写数据
        :param params: 这个参数里面一定要有crawl.
        :return:
        """
        if params is None or len(params) == 0 or ("site_" not in params) or ("projectCode_" not in params) or (
                "code_" not in params):
            # 不存在crawl的信息,没法执行
            return ResultVO(FAIL_CODE, "执行失败,没有Crawl的信息")

        # 找到对应的dataCrawl
        data_crawl = DataCrawlClient.get(params["site_"], params["projectCode_"], params["code_"])

        if data_crawl is None or data_crawl.type != "CUSTOM":
            return ResultVO(FAIL_CODE, f"执行失败,没有找到Code:{params['code_']}的Crawl,或者该Crawl类型不为Custom")

        # 每一次新的任务都重置这个标识
        self.killed = False

        try:
            self.set_local_code(data_crawl.site + "_" + data_crawl.projectCode + "_" + data_crawl.code)
            self.set_process_id(params.get("__processLogId__"))
            self.__task_log_id_local.key = params.get("__logId__")
            self.__execution_message_local.content = []
            TaskLogFileAppender.prepare_to_log(datetime.strptime(params["__triggerTime__"], "%Y-%m-%d %H:%M:%S"),params["__logId__"])
            result_list = self.do_crawl(params)
        except Exception as e:
            import traceback
            return ResultVO(FAIL_CODE, f"执行失败,Code:{params['site_']}的Crawl,发生错误:\n{traceback.format_exc()}")
        finally:
            # 清理 MessageLocal 和 LocalCode
            self.clear_local_code()
            self.clear_process_id()
            self.__task_log_id_local.key = None
            self.__execution_message_local.content = None
            TaskLogFileAppender.end_log()

        message = "success" if self.__execution_message_local.content is None or len(
            self.__execution_message_local.content) == 0 else str(self.__execution_message_local.content)

        if result_list is None or len(result_list) == 0:
            # 搞完了,没的结果,不处理
            return ResultVO(SUCCESS_CODE, message)

        # 有结果,就需要从crawl的配置中找到目标的fd,然后调用fd的接口进行保存
        output_fd_codes = data_crawl.outputFdCodes

        insert_cache = defaultdict(list)
        update_cache = defaultdict(list)
        upsert_cache = defaultdict(list)

        for serializableOutputWrapper in result_list:
            output_fd = self.find_output_fd(output_fd_codes, serializableOutputWrapper.index())
            self.put_in_cache(data_crawl.code, insert_cache, update_cache, upsert_cache, output_fd,
                              serializableOutputWrapper)

        try:
            TaskLogFileAppender.prepare_to_log(datetime.strptime(params["__triggerTime__"], "%Y-%m-%d %H:%M:%S"),
                                               params["__logId__"])
            self.flush_cache(data_crawl.code, insert_cache, update_cache, upsert_cache)
        finally:
            TaskLogFileAppender.end_log()

        return ResultVO(SUCCESS_CODE, message)

    def kill(self):
        """
        由于TaskThread就保证了 同一个ITaskHandler同一时间只能为一个logId服务,因此这里可以直接设置一个成员变量即可,不会有问题
        :return:
        """
        self.killed = True

    def is_killed(self):
        return self.killed

    def get_execution_message_cache(self):
        """
        返回message_cache
        :return:
        """
        return self.__execution_message_local.content

    def get_code_thread_local(self):
        """
        实现上下文的code
        :return:
        """
        return self.__code_local_host

    def get_process_thread_local(self):
        """
        上下文的Id
        :return:
        """
        return self.__process_id_local

    def get_task_log_id(self):
        return self.__task_log_id_local

    @abstractmethod
    def do_crawl(self, params: dict) -> list:
        """
        真正执行数据挖掘的逻辑
        :param params: 需要存入fd中的数据
        :return:
        """



from abc import abstractmethod

from rule_engine import RuleFacts, IRuleListener


class IAgenda:
    """
    规则冲突集.也就是这个地方做匹配
    """

    @abstractmethod
    def get_matched_rules(self, ruleFacts:RuleFacts)->list:
        """
        传入实例,获取匹配的条件
        :param ruleFacts:
        :return:
        """
        pass


class IRuleGroupSession:
    """
    规则组的实例
    """

    @abstractmethod
    def add_facts(self, ruleFacts:RuleFacts):
        """
        添加运行的事实,也就是条件的实例
        :param ruleFacts:
        :return:
        """
        pass

    @abstractmethod
    def add_fact(self, key:str, value):
        """
        添加运行的事实,也就是条件的实例
        :param key:
        :param value:
        :return:
        """
        pass

    @abstractmethod
    def add_rule_listener(self, rule_listener:IRuleListener):
        """
        增加规则执行的消息监听
        :param rule_listener:
        :return:
        """
        pass

    @abstractmethod
    def execute(self)->list:
        """
        执行规则对比,得到结果
        :return:
        """
        pass


from enum import Enum

_defaultValues = {'STRING': "''", 'INTEGER': '0', 'LONG': '0', 'FLOAT': '0', 'DOUBLE': '0', 'BOOLEAN': "''",
                  'DATE': 'CURRENT_TIMESTAMP()', 'CHAR': "''", 'BYTE': "''", 'MODEL': "''"}


class DataTypeEnum(Enum):
    STRING = "STRING"
    INTEGER = "INTEGER"
    LONG = "LONG"
    FLOAT = "FLOAT"
    DOUBLE = "DOUBLE"
    BOOLEAN = "BOOLEAN"
    DATE = "DATE"
    CHAR = "CHAR"
    BYTE = "BYTE"
    MODEL = "MODEL"

    def __new__(cls, name):
        obj = object.__new__(cls)
        obj._value_ = name
        obj._default_ = _defaultValues.get(name)
        return obj

    def default_value(self):
        return self._default_


class MetaModelField:
    def __init__(self, id=None, modelId=None, fieldName=None, alias=None, description=None,
                 dataType=None,
                 primaryKey=None, minLength=None, maxLength=None,
                 nullable=True, refModel=None, multiple=None, indexNames=None, fixItemId=None, params=None):
        self.id = id
        self.modelId = modelId
        self.fieldName = fieldName
        self.alias = alias
        self.description = description
        self.dataType: DataTypeEnum = DataTypeEnum(dataType)
        self.primaryKey = primaryKey
        self.minLength = minLength
        self.maxLength = maxLength
        self.nullable = nullable
        self.refModel = refModel
        self.multiple = multiple
        self.indexNames = indexNames
        self.fixItemId = fixItemId
        self.params = params

#
# if __name__ == '__main__':
#     a = DataTypeEnum("INTEGER")
#     print(f"{a} 默认值:{a.default_value()}")
#
#     b = DataTypeEnum('FLOAT')
#     print(f"{b} 默认值:{b.default_value()}")
#
#     c = DataTypeEnum('DATE')
#     print(f"{c} 默认值:{c.default_value()}")


from api.exceptions import IllegalArgumentException
from rule_engine import IRuleAction, IRuleConditionOperation

__RULE_ENGINE_CUSTOM_CACHE__ = {}


def _is_subclass(v,clazz)->bool:
    cc = None
    if clazz=="IRuleAction":
        cc = IRuleAction
    elif clazz=="IRuleConditionOperation":
        cc = IRuleConditionOperation

    if cc is None:
        return False
    return issubclass(v,cc)

def _find_class(context:dict,clazz):
    script_class = None
    for k, v in context.items():
        if k != clazz and isinstance(v, type) and _is_subclass(v,clazz):
            script_class = v
    return script_class


def load_instance(code_source: str, clazz,name: str=None):
    if name is not None and  name in __RULE_ENGINE_CUSTOM_CACHE__:
        # 说明在缓存中,直接返回
        return __RULE_ENGINE_CUSTOM_CACHE__[name]
    # 说明还没有实例化,
    if code_source is None or len(code_source) == 0:
        raise IllegalArgumentException(f"加载{name or '代码'} 失败,源码为空")
    else:
        # 说明是有东西的,开始尝试加载
        context = {}
        exec(code_source, context)  # None

        # 找到类定义
        script_class = _find_class(context, clazz)

        if  script_class is None:
            # 说明没有找到,报错
            raise IllegalArgumentException(f"加载{name or code_source} 源码失败,没有继承自IRuleAction/IRuleConditionOperation")

        # 放入缓存
        if name is not None:
            __RULE_ENGINE_CUSTOM_CACHE__[name] = script_class()
            # 返回结果
            return __RULE_ENGINE_CUSTOM_CACHE__[name]
        else:
            return script_class()




# -*- coding: utf-8 -*-
import socket
import re
import base64
import gzip

try:
    import urllib.request as urllib2
    from urllib.error import HTTPError
    from urllib.error import URLError
except ImportError:
    import urllib2
    from urllib2 import HTTPError
    from urllib2 import URLError
    from StringIO import StringIO

"""
Default encoding
"""
_DEFAULT_ENCODING = "utf-8"

_URL_REGEX = re.compile(
    r'^((?:http)s?)://'  # http:// or https://
    r'(([A-Z0-9_~!.%]+):([A-Z0-9_~!.%]+)@)?'  # basic authentication -> username:password@
    r'((?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
    r'(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?)|'  # domain name without `.`
    r"(?:\[((?:[A-F0-9]{1,4}:){7}[A-F0-9]{1,4})\])|"  # ipv6
    r'localhost|'  # localhost...
    r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
    r'(?::(\d+))?'  # optional port
    r'(?:/?|[/?]\S+)$', re.IGNORECASE)


def parse_url(url):
    m = _URL_REGEX.match(url)
    if m:
        addr = url
        if m.group(2) is not None:
            addr = addr.replace(m.group(2), "")
            ori_auth = ("%s:%s" % (m.group(3), m.group(4))).encode()
            auth_str = base64.standard_b64encode(ori_auth).decode()
        else:
            auth_str = None
        return {
            "url": addr,
            "auth": auth_str,
            "schema": m.group(1),
            "host": m.group(5),
            "ipv6": m.group(6),
            "port": int(m.group(7)) if m.group(7) is not None else None
        }
    else:
        raise URLError("url[%s] is not a valid url." % url)


class Request(urllib2.Request, object):

    def __init__(self, url, data=None, headers={},
                 origin_req_host=None, unverifiable=False,
                 method=None):
        url_match = _URL_REGEX.match(url)
        if url_match is None:
            raise URLError("Unvalid URL")
        url_obj = parse_url(url)
        url_addr = url_obj["url"]
        url_auth = url_obj["auth"]
        try:
            super(Request, self).__init__(url_addr, data=data, headers=headers,
                                          origin_req_host=origin_req_host, unverifiable=unverifiable,
                                          method=method)
        except TypeError:
            super(Request, self).__init__(url_addr, data=data, headers=headers,
                                          origin_req_host=origin_req_host, unverifiable=unverifiable)
            self.get_method = lambda: method if method is not None else "GET"
        if url_auth is not None:
            self.add_header('Authorization', 'Basic %s' % url_auth)


class HttpClient(object):

    def __init__(self, request=None, data=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
                 cafile=None, capath=None, cadefault=False, context=None):
        self.request = request
        self.data = data
        self.timeout = timeout
        self.cafile = cafile
        self.capath = capath
        self.cadefault = cadefault
        self.context = context

    def urlopen(self):
        res = urllib2.urlopen(self.request, data=self.data, timeout=self.timeout,
                              cafile=self.cafile, capath=self.capath,
                              cadefault=self.cadefault, context=self.context)

        if res.info().get("Content-Encoding") == "gzip":
            try:
                # python2
                f = gzip.GzipFile(fileobj=StringIO(res.read()))
            except NameError:
                f = gzip.GzipFile(fileobj=res)
        else:
            f = res

        txt = f.read().decode(_DEFAULT_ENCODING)
        f.close()
        return txt


__HTTP_CLIENT_CLASS__ = HttpClient


def set_http_client_class(clz):
    assert issubclass(clz, HttpClient)
    global __HTTP_CLIENT_CLASS__
    __HTTP_CLIENT_CLASS__ = clz


def load(url, data=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
         cafile=None, capath=None, cadefault=False, context=None):
    if isinstance(url, urllib2.Request):
        request = url
    elif isinstance(url, str):
        request = Request(url)
    else:
        raise URLError("Unvalid URL")
    request.add_header("Accept-encoding", "gzip")
    return __HTTP_CLIENT_CLASS__(request=request, data=data, timeout=timeout,
                                 cafile=cafile, capath=capath, cadefault=cadefault, context=context).urlopen()



"""
日志记录相关的操作
"""
import requests
from munch import Munch

from api.exceptions import ServiceInnerException, AuthException
from client import check_instance_avaliable, do_service

__all__ = ["log_debug", "log_info", "log_warn", "log_error"]


def log_debug(message):
    """
    打印debug日志
    :param message: 日志记录
    :return:
    """
    if message is None:
        return

    check_instance_avaliable()

    _inner_request(message, "debug")


def log_info(message):
    """
    打印info日志
    :param message: 日志记录
    :return:
    """
    if message is None:
        return

    check_instance_avaliable()

    _inner_request(message, "info")


def log_warn(message):
    """
    打印warn日志
    :param message: 日志记录
    :return:
    """
    if message is None:
        return

    check_instance_avaliable()

    _inner_request(message, "warn")


def log_error(message):
    """
    打印error日志
    :param message: 日志记录
    :return:
    """
    if message is None:
        return

    check_instance_avaliable()

    _inner_request(message, "error")


def _inner_request(message, level):
    """
    构造请求,并发送
    :param message: 消息
    :param level: 级别
    :return:
    """
    request = None
    if not isinstance(message, Munch):
        # 说明不是 Munch类型的 判断是否是Dict的
        if isinstance(message, dict):
            # 如果是dict类型的,直接发
            request = message
    else:
        # 说明是munch的,那么就转成Dict的
        request = message.toDict()
    try:
        r = do_service(f"remote/hl-service/{level}",data=request,method="post",return_type="none")
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e




class ClassInfo(dict):
    @property
    def className(self):
        return self.__getitem__("className")

    @className.setter
    def className(self, value: str):
        self.__setitem__("className", value)

    @property
    def methodName(self):
        return self.__getitem__("methodName")

    @methodName.setter
    def methodName(self, value: str):
        self.__setitem__("methodName", value)

    @property
    def lineNumber(self):
        return self.__getitem__("lineNumber")

    @lineNumber.setter
    def lineNumber(self, value: int):
        self.__setitem__("lineNumber", value)

class LogMessage(dict):

    @property
    def actorName(self):
        return self.__getitem__("actorName")

    @actorName.setter
    def actorName(self, value: str):
        self.__setitem__("actorName", value)

    @property
    def classInfo(self)->ClassInfo:
        return self.__getitem__("classInfo")

    @classInfo.setter
    def classInfo(self, value: ClassInfo):
        self.__setitem__("classInfo", value)

    @property
    def threadName(self):
        return self.__getitem__("threadName")

    @threadName.setter
    def threadName(self, value: str):
        self.__setitem__("threadName", value)

    @property
    def message(self):
        return self.__getitem__("message")

    @message.setter
    def message(self, value: str):
        self.__setitem__("message", value)

    @property
    def traceId(self):
        return self.__getitem__("traceId")

    @traceId.setter
    def traceId(self, value: str):
        self.__setitem__("traceId", value)

    @property
    def exception(self)->Exception:
        return self.__getitem__("exception")

    @exception.setter
    def exception(self, value: Exception):
        self.__setitem__("exception", str(value))


from abc import abstractmethod
from datetime import datetime
from queue import Queue, Empty

from api.handler.outputer.IBaseOutputer import IBaseOutputer
from client import ServiceInnerException
from factdatasource import FactDatasourceProxyService
from schedule.utils.log import FrameworkHandlerLogger


class IBatchOutputer(IBaseOutputer):

    @abstractmethod
    def get_batch_size(self) -> int:
        """
        获取一次批量任务数量
        :return:
        """

    def __logResult(self,task_logger,results):
        if results is not None and len(results) > 0:
            results = filter(lambda x: not x.success, results)
            for res in results:
                task_logger.log(f"{res.error} data:{res.content}")

    def batch_output(self, code, output_fd_codes, is_finished, output_stream: Queue, params):
        """
        批量发送任务出去
        :param params:
        :param code:
        :param output_fd_codes:
        :param is_finished:
        :param output_stream:
        :return:
        """
        insert_cache = {}
        update_cache = {}
        upsert_cache = {}

        from schedule.utils.log import TaskLogFileAppender
        from schedule.utils.log.TaskLogger import TaskLogger
        try:
            task_logger=TaskLogger(TaskLogFileAppender.prepare_to_log(datetime.strptime(params["__triggerTime__"], "%Y-%m-%d %H:%M:%S"),params["__logId__"]))

            while not is_finished[0] or not output_stream.empty():

                if self.is_killed():
                    FrameworkHandlerLogger.log_warn(f"{code} 任务被强制停止,取消数据输出")
                    return

                try:
                    poll = output_stream.get(timeout=1)
                except Empty:
                    # 表示这个任务为null,不处理
                    continue

                output_fd = self.find_output_fd(output_fd_codes, poll.index())
                # 把输出放入缓存中
                self.put_in_cache(code, insert_cache, update_cache, upsert_cache, output_fd, poll)

                # 清空批量任务
                try:
                    serializables = insert_cache.get(output_fd)
                    if serializables is not None and len(serializables) >= self.get_batch_size():
                        results = FactDatasourceProxyService.fd_client_proxy().inserts(output_fd, serializables)
                        self.__logResult(task_logger,results)
                        serializables.clear()

                    serializables = update_cache.get(output_fd)
                    if serializables is not None and len(serializables) >= self.get_batch_size():
                        results = FactDatasourceProxyService.fd_client_proxy().updates(output_fd, serializables)
                        self.__logResult(task_logger, results)
                        serializables.clear()

                    serializables = upsert_cache.get(output_fd)
                    if serializables is not None and len(serializables) >= self.get_batch_size():
                        results = FactDatasourceProxyService.fd_client_proxy().save_or_updates(output_fd, serializables)
                        self.__logResult(task_logger, results)
                        serializables.clear()
                except Exception as e:
                    import traceback
                    raise ServiceInnerException(f"执行失败,Code: {code}的执行器,保存数据发生错误:\n{traceback.format_exc()}")

            # 把剩余的任务清空了
            self.flush_cache(code, insert_cache, update_cache, upsert_cache)
        finally:
            TaskLogFileAppender.end_log()

        # 到这来了 就表示是完成了
        return



from api.model.MetaModelField import MetaModelField


class MetaModel:

    @property
    def dao(self):
        from factdatasource.dao.jdbc.JdbcDatasourceDao import JdbcDatasourceDao
        return JdbcDatasourceDao.instance()

    """
    模型
    """

    def __init__(self, id=None, site=None, project_code=None, code=None, name=None, description=None, fields=None):
        """
        数据模型
        :param id:
        :param site:
        :param project_code:
        :param code:
        :param name:
        :param description:
        :param fields: MetaModelField
        """
        self.id = id
        self.site = site
        self.projectCode = project_code
        self.code = code
        self.name = name
        self.description = description
        self.fields = fields

        if self.site is None and self.id is not None:
            # 说明只传入了一个ID,需要获取metamodel
            self.__init_model__(id)

    def __init_model__(self, id):
        sql = f"""SELECT id,site,projectCode,code,name,description FROM `sc_metamodel` WHERE id=:id"""
        code_dict = {'id': id}
        model = self.dao.select_one(sql, code_dict)

        if model is not None:
            self.site = model["site"]
            self.projectCode = model["projectCode"]
            self.code = model["code"]
            self.name = model["name"]
            self.description = model["description"]

            self.fields = self.__init_field__(id)

    def __init_field__(self, id):
        """
        构造 Fields
        :param id:
        :return:
        """
        sql = f"""SELECT id,modelId,fieldName,alias,description,dataType,primaryKey,minLength,maxLength,nullable,
        refModel,multiple,indexNames,fixItemId,params
         FROM `sc_metamodel_field` WHERE modelId=:id"""
        fields = self.dao.select_all(sql, {"id": id})

        if fields is not None and len(fields) > 0:
            result = []
            for field in fields:
                result.append(MetaModelField(**field))
            return result

        return None



from celery import Celery

import config
from config import annotation

celery = Celery('tasks')
celery.config_from_object('cloudcelery.celery_config')
celery.conf.timezone = 'Asia/Shanghai'
celery.conf.setdefault('CELERY_DEFAULT_QUEUE', config.actor_name)  # 关心的是actor_name类型的任务

try:
    concurrent_count = annotation.get_value("${celery.concurrency:1}")
except:
    concurrent_count = 1
celery.conf.setdefault('CELERYD_CONCURRENCY', concurrent_count)  # 设置任务并发数,默认是1

redis_url: str = annotation.get_value("${redis.url:}")

if redis_url is None or redis_url.strip() == '':
    # 说明是新的格式
    database = annotation.get_value("${spring.data.redis.database:0}")
    password = annotation.get_value("${spring.data.redis.password:}")
    nodes = annotation.get_value("${spring.data.redis.sentinel.nodes:}")
    if nodes is None or nodes == '':
        # 说明是单机
        host = annotation.get_value("${spring.data.redis.host:}")
        # 说明是单机
        url = f'redis://{f":{password}@" if (password is not None and password != "") else ""}{host}/{database}'
    else:
        # 说明是哨兵
        master = annotation.get_value("${spring.data.redis.sentinel.master:}")
        split = nodes.split(",")
        url = ";".join(
            [
                f'sentinel://{f":{password}@" if (password is not None and password != "") else ""}{url}/{database}'
                for
                url in split])
        celery.conf.broker_transport_options = {'master_name': master}
        celery.conf.result_backend_transport_options = {'master_name': master}
else:
    try:
        redis_password = annotation.get_value("${redis.password:}")
    except:
        redis_password = None

    try:
        redis_database = annotation.get_value("${redis.database:0}")
    except:
        redis_database = 0

    # 选择redis类型的缓存时,redis服务器的地址 格式为: ip:port
    # 现在支持了redis3-cluster,多个IP使用逗号分割
    # 现在支持了redis3-sentinel哨兵模式,使用分号分割. 第一个表示主节点名字. MASTERNAME;哨兵IP1:PORT1;哨兵IP2:PORT2;哨兵IP3:PORT3的方式
    if "," in redis_url:
        # 说明是redis-cluster 那么就要初始化集群
        raise NotImplementedError("celery还不支持,see: https://github.com/celery/kombu/pull/1021")
    elif ";" in redis_url:
        # 说明是哨兵形式的主从集群
        # app.conf.broker_url = 'sentinel://localhost:26379;sentinel://localhost:26380;sentinel://localhost:26381'
        # app.conf.broker_transport_options = {'master_name': "cluster1"}
        split = redis_url.split(";")
        sentinel = split[0]
        split.remove(sentinel)
        url = ";".join(
            [
                f'sentinel://{f":{redis_password}@" if (redis_password is not None and redis_password != "") else ""}{url}/{redis_database}'
                for
                url in split])
        celery.conf.broker_transport_options = {'master_name': sentinel}
        celery.conf.result_backend_transport_options = {'master_name': sentinel}
    else:
        # 说明是单机
        url = f'redis://{f":{redis_password}@" if (redis_password is not None and redis_password != "") else ""}{redis_url}/{redis_database}'

celery.conf.broker_url = url  # 中间件
celery.conf.result_backend = url  # 结果存储

# 这里为了监控,需要再连接一个redis
from redis import Redis


def celery_redis_client() -> Redis:
    return celery.backend.client



from abc import abstractmethod

from api.exceptions import IllegalArgumentException
from api.handler.ICacheAble import ICacheAble
from client import ScheduleCacheClient

CACHE_PREFIX = "sobeyficus.cache."
PROCESS_CACHE_PREFIX = "sobeyficus.process.cache."


class ICacheAbleHandler(ICacheAble):
    """
    缓存操作的接口
    """

    @abstractmethod
    def get_code_thread_local(self):
        """
        需要实现的方法,返回一个 thread_local对象
        :return:
        """

    @abstractmethod
    def get_process_thread_local(self):
        """
        需要实现的方法,返回一个 thread_local对象
        :return:
        """

    # region 同执行器的缓存
    def set_cache_value(self, key, value):
        """
        放入缓存
        :param key:
        :param value:
        :return:
        """
        if key is None or value is None:
            raise IllegalArgumentException("设置任务缓存失败,传入的参数不合法")
        ScheduleCacheClient.set_value(self.__generate_key(key), value)

    def set_cache_value_if_absent(self, key, value):
        """
        放入缓存
        :param key:
        :param value:
        :return:
        """
        if key is None or value is None:
            raise IllegalArgumentException("设置任务缓存失败,传入的参数不合法")
        return ScheduleCacheClient.set_if_absent(self.__generate_key(key), value)

    def get_cache_value(self, key):
        """
        获取缓存
        :param key:
        :return:
        """
        if key is None:
            raise IllegalArgumentException("获取任务缓存失败,传入的参数不合法")
        return ScheduleCacheClient.get(self.__generate_key(key))

    def delete_cache_value(self, key):
        """
        删除缓存
        :param key:
        :return:
        """
        if key is None:
            raise IllegalArgumentException("删除任务缓存失败,传入的参数不合法")
        ScheduleCacheClient.delete(self.__generate_key(key))
    # endregion

    # region 同执行计划的缓存
    def set_cache_value_from_process(self, key, value):
        """
        放入缓存
        :param key:
        :param value:
        :return:
        """
        if key is None or value is None:
            raise IllegalArgumentException("设置任务缓存失败,传入的参数不合法")

        tmp = self.__generate_key_4_process(key)

        if tmp is None:
            raise IllegalArgumentException("设置执行计划缓存失败,没有找到执行计划的Id")

        ScheduleCacheClient.set_value(tmp, value)

    def set_cache_value_if_absent_from_process(self, key, value):
        """
        放入缓存
        :param key:
        :param value:
        :return:
        """
        if key is None or value is None:
            raise IllegalArgumentException("设置任务缓存失败,传入的参数不合法")

        tmp = self.__generate_key_4_process(key)

        if tmp is None:
            raise IllegalArgumentException("设置执行计划缓存失败,没有找到执行计划的Id")

        return ScheduleCacheClient.set_if_absent(tmp, value)

    def get_cache_value_from_process(self, key):
        """
        获取缓存
        :param key:
        :return:
        """
        if key is None:
            raise IllegalArgumentException("获取任务缓存失败,传入的参数不合法")

        tmp = self.__generate_key_4_process(key)

        if tmp is None:
            raise IllegalArgumentException("设置执行计划缓存失败,没有找到执行计划的Id")

        return ScheduleCacheClient.get(tmp)

    def delete_cache_value_from_process(self, key):
        """
        删除缓存
        :param key:
        :return:
        """
        if key is None:
            raise IllegalArgumentException("删除任务缓存失败,传入的参数不合法")

        tmp = self.__generate_key_4_process(key)

        if tmp is None:
            raise IllegalArgumentException("设置执行计划缓存失败,没有找到执行计划的Id")

        ScheduleCacheClient.delete(tmp)
    # endregion

    def __generate_key(self, key):
        """
        生成key
        :param key:
        :return:
        """
        return CACHE_PREFIX + self.get_code_thread_local().key + "." + key

    def __generate_key_4_process(self, key):
        """
        生成key
        :param key:
        :return:
        """
        try:
            if self.get_process_thread_local().key is None:
                return None

            return PROCESS_CACHE_PREFIX + self.get_process_thread_local().key + "." + key
        except:
            # 有可能 self.get_process_thread_local().key 为空,那就返回None
            return None

    # region ThreadLocal Code相关
    def set_local_code(self, code):
        """
        设置key
        :param code:
        :return:
        """
        self.get_code_thread_local().key = code

    def clear_local_code(self):
        """
        清空key
        :return:
        """
        self.get_code_thread_local().key = None

    def set_process_id(self,code):
        """
        设置Key
        :param code:
        :return:
        """
        self.get_process_thread_local().key = code

    def clear_process_id(self):
        """
        清空key
        :return:
        """
        self.get_process_thread_local().key = None
    # endregion


from api.exceptions import UnSupportException
from rule_engine import IRuleConditionOperation

def compare_equal(lhs,rhs):
    if lhs is rhs:
        return True

    if lhs is None or rhs is None:
        return False

    return lhs == rhs
    # python 直接用==就可以比较 dict list obj 这些
    # if not isinstance(lhs, list):
    #     return lhs == rhs
    # else:
    #     if len(lhs) != len(rhs):
    #         return False
    #     for index, content in enumerate(lhs):
    #         if not  __compare_equal(content,rhs[index]):
    #             return False
    #     return True

class DoNothingOperation(IRuleConditionOperation):
    """
    没用的,永远返回true
    """
    def do_operation(self, factValue, targetValue) -> bool:
        return True

class UnSupportOperation(IRuleConditionOperation):

    def do_operation(self, factValue, targetValue) -> bool:
        raise UnSupportException("此条件动作Python 并不支持")

class DefaultEqualOperation(IRuleConditionOperation):
    def do_operation(self, factValue, targetValue) -> bool:
        # 相等判断
        return compare_equal(factValue,targetValue)

class DefaultNotEqualOperation(IRuleConditionOperation):
    def do_operation(self, factValue, targetValue) -> bool:
        # 相等判断
        return not compare_equal(factValue,targetValue)

class DefaultGTEOperation(IRuleConditionOperation):
    def do_operation(self, factValue, targetValue) -> bool:
        if factValue is None or targetValue is None:
            return False

        try:
            return factValue >= targetValue
        except TypeError:
            raise UnSupportException("不支持 GTE操作")

class DefaultGTOperation(IRuleConditionOperation):
    def do_operation(self, factValue, targetValue) -> bool:
        if factValue is None or targetValue is None:
            return False

        try:
            return factValue > targetValue
        except TypeError:
            raise UnSupportException("不支持 GT操作")

class DefaultLTEOperation(IRuleConditionOperation):
    def do_operation(self, factValue, targetValue) -> bool:
        if factValue is None or targetValue is None:
            return False

        try:
            return factValue <= targetValue
        except TypeError:
            raise UnSupportException("不支持 LTE操作")

class DefaultLTOperation(IRuleConditionOperation):
    def do_operation(self, factValue, targetValue) -> bool:
        if factValue is None or targetValue is None:
            return False

        try:
            return factValue < targetValue
        except TypeError:
            raise UnSupportException("不支持 LT操作")

class DefaultINOperation(IRuleConditionOperation):
    def do_operation(self, factValue, targetValue) -> bool:
        if targetValue is None:
            return False

        if not isinstance(targetValue, list):
            raise UnSupportException("比对目标值并不是一个List,无法对比")

        try:
            return factValue in targetValue
        except TypeError:
            raise UnSupportException("比对目标值并不是一个List,无法对比")

class DefaultRangeOperation(IRuleConditionOperation):
    """
    前开后闭
    """
    def do_operation(self, factValue, targetValue) -> bool:
        if targetValue is None:
            return False

        if not isinstance(targetValue, list) or len(targetValue)!=2:
            raise UnSupportException("比对目标值必须是一个二元数组[start,end]")

        try:
            return factValue >= targetValue[0] and factValue < targetValue[1]
        except TypeError:
            raise UnSupportException("不支持 LT操作")

class PyExpOperation(IRuleConditionOperation):
    """
    python表达式的方式,类似于java中的mvel
    """
    def do_operation(self, factValue, targetValue) -> bool:

        context = factValue or {}

        from munch import Munch
        context = Munch.fromDict(context)
        params = Munch.fromDict(context)

        exec(f"def __b(params):\n{self.__reindent(targetValue,1)}\n", context)
        result2 = context["__b"]
        ar = result2(params)

        return ar if ar is not None else False

    def __reindent(self,s, numSpaces):
        """
        多行代码的统一缩进处理
        :param s:
        :param numSpaces:
        :return:
        """
        leading_space = numSpaces * ' '
        lines = [leading_space + line for line in s.splitlines()]
        return '\n'.join(lines)


from datetime import datetime

from requests.auth import AuthBase

from config.annotation import Value


# region 从配置文件中读取是否开启认证的配置
@Value("${sobeycube.oauth.enable:none}")
def oauth2_type():
    pass


@Value("${sobeycube.oauth.client_id:}")
def application_key():
    pass


@Value("${sobeycube.oauth.client_secret:}")
def secret_key():
    pass

@Value("${sobeycube.oauth.oauth_uri}")
def user_center_uri():
    pass

@Value("${sobeycube.oauth.username:}")
def user_name():
    pass


@Value("${sobeycube.oauth.password:}")
def user_password():
    pass


# endregion

# region oauth2的认证
access_token = None
oauth_expiration: datetime = None


def acquire_access_token():
    """
    重新申请accessToken
    :return:
    """
    from client import AuthClient

    server, grant_type = ("sobeyficus", "client_credentials") if oauth2_type() == "server" else (
        user_center_uri(), "password")

    token, expiration = AuthClient.oauth_access_token(server, grant_type, application_key(), secret_key(), user_name(),
                                                      user_password())

    global access_token
    access_token = token

    global oauth_expiration
    oauth_expiration = expiration

    return access_token


def get_oauth_token():
    """
    生成oauth2的token
    :return:
    """

    global access_token
    if access_token is None or (oauth_expiration is not None and oauth_expiration <= datetime.now()):
        #  说明没有token, 或者token过期了
        access_token = acquire_access_token()

    return access_token


def _extract_auth_str(tokenType):
    """
    返回oauth的token
    :param tokenType:
    :return:
    """
    return f"{tokenType} {get_oauth_token()}"


# endregion

# region JWT的认证
jwt_token = None
jwt_expiration: datetime = None


def acquire_jwt_token():
    """
    重新申请jwtToken
    :return:
    """
    from client import AuthClient
    token, expiration = AuthClient.oauth_jwt_token(user_name(), user_password())
    jwt_token = token

    global jwt_expiration
    jwt_expiration = expiration

    return jwt_token


def _extract_jwt_str():
    global jwt_token
    if jwt_token is None or (jwt_expiration is not None and jwt_expiration <= datetime.now()):
        #  说明没有token, 或者token过期了
        jwt_token = acquire_jwt_token()

    return jwt_token


# endregion

_oauth2_type = None


class ClientAuth(AuthBase):
    """
    自定义的client认证
    """

    def __init__(self):
        global _oauth2_type
        if _oauth2_type is None:
            # 需要加载认证类型
            _oauth2_type = oauth2_type()

    def __call__(self, r):
        """
        必须实现的这个方法,类似于JAVA中的 filter. 在调用的的时候增加认证的东西
        :param r: PreparedRequest实例
        :return:
        """

        if _oauth2_type == "none" or _oauth2_type == False:
            # 如果没有打开认证,就直接返回
            return r

        # 如果开启了认证,需要判断是哪一种认证方式,server?client?
        # 如果是server认证的,就需要提供两个认证,一个是oauth2,一个是jwt的

        # 先校验oauth2的
        r.headers['Authorization'] = _extract_auth_str("Bearer")

        if _oauth2_type == "server" or _oauth2_type == True:
            # 再处理jwt的认证
            if user_name() is not None and user_password() is not None:
                # 开启了认证,并且填写了userName和password.才增加这个请求头
                r.headers['sobeycube-http-token'] = _extract_jwt_str()

        return r



# 规则引擎
from readerwriterlock import rwlock

from rule_engine import RuleGroupInfo

# 缓存, key是 project_name
from rule_engine.core import IRuleGroupSession
from rule_engine.core.SimpleRuleGroupSession import SimpleRuleGroupSession

__cache = {}

__read_write_lock = rwlock.RWLockRead()

def __groupKey(name,project):
    return f"{project}_{name}"

def register_rule_group(rule_group_info:RuleGroupInfo):
    """
    增加一个规则组,如果原来的存在,就替换
    :arg rule_group_info 规则
    :return:
    """
    key = __groupKey(rule_group_info.name,rule_group_info.project)

    with __read_write_lock.gen_wlock():
        __cache[key] = rule_group_info

def unregister_rule_group(project:str, name:str):
    """
    删除掉一个规则组
    :param project:
    :param name:
    :return:
    """
    key = __groupKey(name,project)
    with __read_write_lock.gen_wlock():
        del __cache[key]

def clear_rule_group(project:str):
    """
    清除一个规则组
    :param project:
    :return:
    """
    with __read_write_lock.gen_wlock():
        for k in list(__cache.keys()):
            if k.startswith(f"{project}_"):
                del __cache[k]


def list_rule_group_names(project:str):
    """
    列出规则组的名字
    :param project:
    :return:
    """
    result = []
    with __read_write_lock.gen_rlock():
        for k in list(__cache.keys()):
            if k.startswith(f"{project}_"):
                result.append(k.replace(f"{project}_",""))

    return result

def create_new_session(group_name:str, project:str)->IRuleGroupSession:
    """
    执行一个规则
    :param group_name:
    :param project:
    :return:
    """
    key = __groupKey(group_name, project)
    with __read_write_lock.gen_rlock():
        if key not in __cache:
            from api.exceptions import IllegalArgumentException
            raise IllegalArgumentException(f"没有找到{project}:{group_name}的规则组")

        rule_group_info = __cache[key]

    return SimpleRuleGroupSession(rule_group_info)


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/22
from api.model.FactDatasource import FactDatasource
from factdatasource.execptions import NotSupportedFDException
from factdatasource.persistence.jdbc.JdbcFactDatasourceContext import JdbcFactDatasourceContext


class Db2JdbcFactDatasourceContext(JdbcFactDatasourceContext):
    """
    TODO 暂时不支持DB2
    """
    def __init__(self, fact_datasource: FactDatasource):
        raise NotSupportedFDException('暂时不支持该FD类型')



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/23
import datetime
import json
import re
import threading
import uuid


from api.model.FactDatasource import JdbcTypeEnum


def get_jdbc_type(url: str) -> JdbcTypeEnum:
    """
    根据url获取JDBC的类型
    :param url:
    :return:
    """
    if not url:
        return None

    if url.startswith('jdbc:derby:') or url.startswith('jdbc:log4jdbc:derby:'):
        return JdbcTypeEnum.DERBY
    elif url.startswith('jdbc:mysql:') or url.startswith('jdbc:cobar:') or url.startswith('jdbc:log4jdbc:mysql:'):
        return JdbcTypeEnum.MYSQL
    elif url.startswith('jdbc:mariadb:'):
        return JdbcTypeEnum.MARIADB
    elif url.startswith('jdbc:oracle:') or url.startswith('jdbc:log4jdbc:oracle:'):
        return JdbcTypeEnum.ORACLE
    elif url.startswith('jdbc:alibaba:oracle:'):
        return JdbcTypeEnum.ALI_ORACLE
    elif url.startswith('jdbc:microsoft:') or url.startswith('jdbc:log4jdbc:microsoft:'):
        return JdbcTypeEnum.SQL_SERVER
    elif url.startswith('jdbc:sqlserver:') or url.startswith('jdbc:log4jdbc:sqlserver:'):
        return JdbcTypeEnum.SQL_SERVER
    elif url.startswith('jdbc:sybase:Tds:') or url.startswith('jdbc:log4jdbc:sybase:'):
        return JdbcTypeEnum.SYBASE
    elif url.startswith('jdbc:jtds:') or url.startswith('jdbc:log4jdbc:jtds:'):
        return JdbcTypeEnum.JTDS
    elif url.startswith('jdbc:fake:') or url.startswith('jdbc:mock:'):
        return JdbcTypeEnum.MOCK
    elif url.startswith('jdbc:postgresql:') or url.startswith('jdbc:log4jdbc:postgresql:'):
        return JdbcTypeEnum.POSTGRESQL
    elif url.startswith('jdbc:hsqldb:') or url.startswith('jdbc:log4jdbc:hsqldb:'):
        return JdbcTypeEnum.HSQL
    elif url.startswith('jdbc:db2:'):
        return JdbcTypeEnum.DB2
    elif url.startswith('jdbc:h2:') or url.startswith('jdbc:log4jdbc:h2:'):
        return JdbcTypeEnum.H2
    elif url.startswith('jdbc:dm:'):
        return JdbcTypeEnum.DM
    elif url.startswith('jdbc:kingbase:'):
        return JdbcTypeEnum.KINGBASE
    elif url.startswith('jdbc:log4jdbc:'):
        return JdbcTypeEnum.LOG4JDBC
    elif url.startswith('jdbc:vernox:'):
        return JdbcTypeEnum.VERNOX
    else:
        return None


class Singleton(object):
    """
    简单实现一个单例基类, 保证通过instance方法取到的对象都是同一个对象
    这里使用这个类不使用模块直接加载的原因是想延迟初始化，使用集中式的FD时，很多本地的FD配置启动时都不需要加载
    """
    _lock = threading.Lock()
    _instance = None

    @classmethod
    def instance(cls):
        if not cls._instance:
            with cls._lock:
                if not cls._instance:
                    cls._instance = cls()
        return cls._instance


def str_placeholder_replace(prefix: str, suffix: str, value: str, value_map: dict):
    """
    字符串占位符替换工具,将前缀和后缀之间的字符串作为key，在value_map中如果存在该key值，就用该key值替换
    如： str_placeholder_replace('${', '}', 'select * from ${table}', {'table':'test'}) 返回select * from test

    :param prefix:
    :param suffix:
    :param value:
    :param value_map:
    :return:
    """
    if not prefix or not suffix or not value or not value_map:
        return value

    value_list = []
    start = 0
    regex = re.escape(prefix) + '|' + re.escape(suffix)
    for r in re.finditer(regex, value):
        fix_key = value[start:r.start()]
        fix = value[r.start():r.end()]
        if fix == prefix:
            # 入栈
            value_list.append(fix_key)
            value_list.append(fix)
        elif fix == suffix:
            k = None
            key = fix_key

            # 出栈
            if value_list:
                k = value_list.pop()
                while value_list and prefix != k:
                    key = k + key
                    k = value_list.pop()

            # 替换值
            if k and k == prefix:
                if key in value_map:
                    v = value_map.get(key)
                    value_list.append(str(v))
                else:
                    value_list.append(prefix + key + suffix)
            else:
                k = k if k else ''
                key = key if key else ''
                value_list.append(k + key + suffix)
        start = r.end()
    # 加上最后剩下的
    if start < len(value):
        value_list.append(value[start:])
    return ''.join(value_list)


class JSONEncoder(json.JSONEncoder):
    """
    dict转换为JSON时处理下
    """
    def default(self, o):
        if isinstance(o, datetime.date):
            return str(o)
        if isinstance(o, uuid.UUID):
            return str(o)

        try:
            from bson import ObjectId
            if isinstance(o, ObjectId):
                return str(o)
        except ImportError:
            pass

        return json.JSONEncoder.default(self, o)


def uuid4():
    return str(uuid.uuid4()).replace('-', '')




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
from abc import abstractmethod

from api.model.Page import Page


class FactDatasourceContext(object):

    @abstractmethod
    def fd(self):
        """
        返回各自保存的fd对象
        :return:
        """

    @abstractmethod
    def size(self) -> int:
        """
        返回数据总长度
        :return: 数据条数:long
        """

    @abstractmethod
    def is_empty(self) -> bool:
        """
        返回是否存在数据
        :return: boolean
        """

    @abstractmethod
    def collect(self, size: int) -> list:
        """
        返回指定条数的数据
        :param size: 返回的条数
        :return: list
        """

    @abstractmethod
    def collect_conditions(self, size: int, condition_groups: list) -> list:
        """
        返回指定条数的数据
        :param size: 返回的条数
        :param condition_groups: 查询条件
        :return: list
        """

    @abstractmethod
    def query(self, query: str, parameters: dict = None) -> Page:
        """
        使用查询语句查询数据
        查询语句query支持${}和#{}这两种参数占位符，其中值放在parameters中，
        在除JDBC以外的FD中${}和#{}两种占位符是没有区别的，都是直接替换，
        在JDBC中是有区别的，JDBC中的${}是直接进行替换的，#{}这种占位符会转换为sqlalchemy中的:key这种占位符，参数会传到sqlalchemy中进行处理。
        比如 'SELECT * FROM `sc_dataproject` WHERE site=#{site} AND ttl=${ttl}'   {'site': 'S1','ttl': '1'}
        这种经过占位符处理后会转为 'SELECT * FROM `sc_dataproject` WHERE site=:site AND ttl=1'

        parameters支持的默认参数：
            'pageNum_': 页数
            'pageSize_': 每页数量
            'needCount_': 是否进行总数查询
        :param query: 查询语句
        :param parameters: 查询参数
        :return: Page
        """

    @abstractmethod
    def inserts(self, result_list: list) -> list:
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        result_list里边存放的应该是一个dict或者一个可序列化的obj，这个dict的key不能以:开头
        :param result_list: 要保存的数据
        :return: None
        """

    @abstractmethod
    def updates(self, result_list: list) -> list:
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的,主键的key要求使用'id'
        result_list里边存放的应该是一个dict或者一个可序列化的obj，这个dict的key不能以:开头
        :param result_list: 要修改的数据
        :return: None
        """

    @abstractmethod
    def inserts_or_updates(self, result_list: list) -> list:
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的,主键的key要求使用'id'
        result_list里边存放的应该是一个dict或者一个可序列化的obj，这个dict的key不能以:开头
        :param result_list: 要添加或者需要修改的数据
        :return: None
        """

    @abstractmethod
    def delete_all(self):
        """
        清空数据
        :return: None
        """

    @abstractmethod
    def delete(self, query: str):
        """
        根据删除语句删除数据,query是完整的删除语句
        :param query: 完整的删除语句
        :return: None
        """

    @abstractmethod
    def delete_conditions(self, condition_groups: list):
        """
        根据删除条件,构造删除语句
        :param condition_groups:
        :return:
        """

    # @abstractmethod
    # def get_fact_datasource_fields(self):
    #     """
    #     获取fd的字段
    #     :return: List<FactDatasourceField>
    #     """



from api.exceptions import IllegalArgumentException
from factdatasource import FactDatasourceProxyService


class FdInputWrapper:
    """
    fd输入的一个包装
    """
    __fd_code = None
    __join_fd_codes = set()

    def fd_code(self):
        return self.__fd_code

    def __init__(self, fd_code):
        self.__fd_code = fd_code
        self.__join_fd_codes.add(fd_code)

    def join(self, other_fd_input):
        """
        合并两个fd的查询
        :param other_fd_input:
        :return:
        """
        # 1.check 两个fd的 链接是不是同一个,只能是同一个的才能链接
        this_fact_datasource = FactDatasourceProxyService.fd_client_proxy().fd(self.__fd_code)
        other_fact_datasource = FactDatasourceProxyService.fd_client_proxy().fd(other_fd_input.fd_code())

        if this_fact_datasource is None or other_fact_datasource is None:
            raise IllegalArgumentException(f"无法联合两个FD({self.__fd_code},{other_fd_input.fd_code()}),无法查询到FD数据")

        if this_fact_datasource.type != other_fact_datasource.type:
            raise IllegalArgumentException(f"无法联合两个FD({self.__fd_code},{other_fd_input.fd_code()}),两个FD的类型不相同")

        if this_fact_datasource.connection != other_fact_datasource.connection:
            raise IllegalArgumentException(f"无法联合两个FD({self.__fd_code},{other_fd_input.fd_code()}),两个FD的连接不相同")

        self.__join_fd_codes.add(other_fd_input.fd_code())
        return self

    def query(self, query, parameters={}):
        """
        对fd进行查询
        :param query: 查询语句
        :param parameters: 上下文参数
        :return: Munch对象数据
        """
        # TODO 这个地方需要对query进行校验.验证他涉及的表是不是都是fd里面的

        if self.__fd_code not in self.__join_fd_codes:
            raise IllegalArgumentException(f"来源FD不存在:{self.fd_code()}")

        return FactDatasourceProxyService.fd_client_proxy().query(self.__fd_code, query, parameters)

    def info(self):
        """
        获取fd的信息
        :return:
        """
        return FactDatasourceProxyService.fd_client_proxy().fd(self.__fd_code)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
import math
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED

from api.exceptions import IllegalArgumentException
from api.model import MetaModelTransform
from api.model.FactDatasource import FactDatasource
from factdatasource.FactDatasourceContext import FactDatasourceContext
from factdatasource.execptions import DatasourceNotFoundException
from factdatasource.transformer import TransformerContext
from libs.utils import str_placeholder_replace


class AbstractFactDatasourceContext(FactDatasourceContext):
    # 默认的分片大小
    default_partition_size = 200
    # 多主键值的连接符
    PRIMARY_CONNECT_STRING = '_'
    # 使用统一的线程池来管理
    __fd_thread_pool = ThreadPoolExecutor(max_workers=20, thread_name_prefix='ficus-fd-')

    def __init__(self, fact_datasource: FactDatasource):
        self.fact_datasource = fact_datasource

    def _fd_placeholder_replace(self, value: str, value_map: dict):
        """
        fd参数的占位符处理
        处理替换${}和#{}两种占位符
        :param value:
        :param value_map:
        :return:
        """
        if not isinstance(value, str):
            raise IllegalArgumentException('占位符处理参数类型错误，value只能为str.')
        if not value_map:
            return value
        value = str_placeholder_replace('${', '}', value, value_map)
        value = str_placeholder_replace('#{', '}', value, value_map)
        return value

    def fd(self) -> FactDatasource:
        if not self.fact_datasource:
            raise DatasourceNotFoundException('未配置数据源')
        return self.fact_datasource

    def inserts(self, result_list: list) -> list:
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param result_list: 要保存的数据
        :return:
        """
        if not result_list:
            return None
        if not isinstance(result_list, list):
            result_list = [result_list]

        table = self.fd().target

        partition = self._compute_and_partition_datas(result_list)

        if len(partition) == 1:
            return self._proxy_single_thread_inserts(table, result_list)
        else:
            all_tasks = [self.__fd_thread_pool.submit(self._proxy_single_thread_inserts, table, data) for data in
                         partition]
            wait(all_tasks, return_when=ALL_COMPLETED)
            result = []
            for task in all_tasks:
                result.extend(task.result())
            return result

    def updates(self, result_list: list):
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要修改的数据
        :return:
        """
        if not result_list:
            return
        if not isinstance(result_list, list):
            result_list = [result_list]

        table = self.fd().target

        partition = self._compute_and_partition_datas(result_list)

        if len(partition) == 1:
            return self._proxy_single_thread_updates(table, result_list)
        else:
            all_tasks = [self.__fd_thread_pool.submit(self._proxy_single_thread_updates, table, data) for data in
                         partition]
            wait(all_tasks, return_when=ALL_COMPLETED)
            result = []
            for task in all_tasks:
                result.extend(task.result())
            return result

    def inserts_or_updates(self, result_list: list):
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        if not result_list:
            return
        if not isinstance(result_list, list):
            result_list = [result_list]

        table = self.fd().target

        partition = self._compute_and_partition_datas(result_list)

        if len(partition) == 1:
            return self._proxy_single_thread_inserts_or_updates(table, result_list)
        else:
            all_tasks = [self.__fd_thread_pool.submit(self._proxy_single_thread_inserts_or_updates, table, data) for
                         data in
                         partition]
            wait(all_tasks, return_when=ALL_COMPLETED)
            result = []
            for task in all_tasks:
                result.extend(task.result())
            return result

    def _proxy_single_thread_inserts(self, table: str, result_list: list) -> list:
        self._execute_model_field_transform(result_list)
        return self._single_thread_inserts(table, result_list)

    @abstractmethod
    def _single_thread_inserts(self, table: str, result_list: list) -> list:
        """单线程insert数据"""

    def _proxy_single_thread_updates(self, table: str, result_list: list) -> list:
        self._execute_model_field_transform(result_list)
        return self._single_thread_updates(table, result_list)

    @abstractmethod
    def _single_thread_updates(self, table: str, result_list: list) -> list:
        """单线程update数据"""

    def _proxy_single_thread_inserts_or_updates(self, table: str, result_list: list) -> list:
        self._execute_model_field_transform(result_list)
        return self._single_thread_inserts_or_updates(table, result_list)

    @abstractmethod
    def _single_thread_inserts_or_updates(self, table: str, result_list: list) -> list:
        """单线程upsert数据"""

    def _compute_and_partition_datas(self, data: list) -> list:
        """
        计算,并尽量均分的分片 数据小于 default_partition_size 就不分片了
        :param data:
        :return: list<result_list>
        """
        if not data:
            return None

        if len(data) <= self.default_partition_size:
            return [data]
        # 共分多少片
        n = math.ceil(len(data) / self.default_partition_size)
        # 每片多少个
        num = math.ceil(len(data) / n)
        return [data[i:i + num] for i in range(0, len(data), num)]

    def _execute_model_field_transform(self, result_list):
        """
        字段转换相关
        :param result_list:
        :return:
        """
        if result_list is None or len(result_list) == 0:
            return

        has_transform_meta_model_fields = self._need_transform_model_field()

        if has_transform_meta_model_fields is None or len(has_transform_meta_model_fields) == 0:
            return

        # 说明是需要进行字段转换的
        for serializable in result_list:
            self._inner_execute_model_field_transform(serializable, has_transform_meta_model_fields)

    def _need_transform_model_field(self):
        """
        判断是否需要进行模型字段的转换
        先每一次都判断一次,以后可能再使用缓存
        :return: Map<MetaModelField,List<MetaModelTransform>>
        """
        model = self.fd().model
        if model is None:
            return None

        fields = model.fields
        if fields is None or len(fields) == 0:
            return None
        result = {}
        for field in fields:
            transforms = MetaModelTransform.init_transform_by_field(field.id)
            if transforms is not None:
                result[field] = transforms

        return result

    def _inner_execute_model_field_transform(self, serializable, has_transform_meta_model_fields):
        """
        转换字段
        :param serializable:
        :param has_transform_meta_model_fields:
        :return:
        """

        if serializable is None:
            return

        if isinstance(serializable, list) or isinstance(serializable, set):
            # array或list类型
            for o in serializable:
                self._inner_execute_model_field_transform(o, has_transform_meta_model_fields)
        else:

            for meta_model_field, transforms in has_transform_meta_model_fields.items():
                # 这里是要写入数据库的值
                value = serializable[meta_model_field.fieldName] if meta_model_field.fieldName in serializable else None

                for transform in transforms:
                    transformerParams: dict = transform.params

                    tmp = {meta_model_field.fieldName: value}

                    # 补充关联的字段
                    if transformerParams is not None and "__referenceFields__" in transformerParams:
                        referenceFields: str = transformerParams["__referenceFields__"]
                        if referenceFields is not None and "" != referenceFields:
                            split = referenceFields.split(",")
                            for s in split:
                                if s == "":
                                    continue
                                tmp[s] = serializable[s] if s in serializable else None

                    # 执行转换
                    transformer = TransformerContext.get_transformer(transform)
                    if transformer is not None:
                        objectMap: dict = transformer.transform_field(tmp, transformerParams, meta_model_field)
                        for key, value in objectMap.items():
                            serializable[key] = value
                            tmp[key] = value

    def get_primary_keys(self) -> list:
        """
        获取主键列表
        1. 通过fd中配置模型获取主键列表
        2. 如果没有设置模型，则默认使用id作为主键
        :return:
        """
        fd = self.fd()
        result = list()
        if fd and fd.model and fd.model.fields:
            for field in fd.model.fields:
                if field.primaryKey == 1:
                    # 这里表示这个字段设置的是主键
                    result.append(field.fieldName)

        if not result:
            # 如果没有配置主键，这里兼容老版本，设置id为默认主键
            result.append('id')
        return result

    def get_primary_values(self, primary_keys: list, values: dict):
        """
        根据主键列表的key返回一个包含所有主键的dict
        :param primary_keys:
        :param values:
        :return:
        """
        if not primary_keys or not values:
            return None

        primary_keys = primary_keys.copy()
        # 存一份小写数据，避免每次都遍历找
        values_lower = None

        primary_dict = dict()
        for key in primary_keys:
            v = values.get(key)
            if v is None:
                if not values_lower:
                    values_lower = {str(k).lower(): v for k, v in values.items()}
                # 没有找到，不区分大小写再找一次
                v = values_lower.get(key.lower())
            if v is None:
                return None
            primary_dict[key] = v
        return primary_dict

    def get_primary_values_key(self, primary_keys: list, values: dict):
        """
        根据主键列表的字典顺序拼接主键，使用下划线连接，定义的所有主键都不能为空，主键key不区分大小写
        :return:
        """
        if not primary_keys or not values:
            return None

        primary_dict = self.get_primary_values(primary_keys, values)
        if not primary_dict:
            return None
        sort_key = sorted(primary_dict)
        primary_values = list()
        for key in sort_key:
            primary_values.append(str(primary_dict.get(key)))
        return self.PRIMARY_CONNECT_STRING.join(primary_values)



SUCCESS_CODE = 3
FAILED_CODE = -1
DOING_CODE = 1

class AsyncServiceResponse:
    """
    异步的响应
    """

    def __init__(self, response_data = None, task_state: int = SUCCESS_CODE, task_desc: str = None,
                 response_kind: str = None):
        """

        :param response_data: 执行响应信息
        :param task_state: 任务状态
        :param task_desc: 任务描述,错误描述
        :param response_kind: 任务种类,可以不填
        """
        self.response_data = response_data
        self.task_state = task_state
        self.task_desc = task_desc
        self.response_kind = response_kind



from flask import jsonify

from . import remote


@remote.route('/actuator/info', methods=['GET'])
def actuator_info():
    return jsonify({"build":{"version":"3.0.1","artifact":"sobeycube-ficus-framework-4-py","name":"FicusFramework","group":"com.sobey.jcg"}})


@remote.route('/actuator/health', methods=['GET'])
def actuator_health():
    return jsonify({"status": "UP"})



SUCCESS_CODE = 2
DOING_CODE = 1
FAIL_CODE = -1


class ResultVO(object):
    code = 2
    msg = "success"
    content = None

    def __init__(self, code=2, msg="success", content=None):
        self.code = code
        self.content = content
        self.msg = msg

    def to_json(self):
        from flask import jsonify
        return jsonify({'code': self.code, 'msg': self.msg, 'content': self.content})

    def to_dict(self):
        return {'code': self.code, 'msg': self.msg, 'content': self.content}


SUCCESS = ResultVO(content="success")
DOING = ResultVO(code=DOING_CODE, msg="doing", content="doing")
FAIL = ResultVO(code=FAIL_CODE, msg="failed", content="failed")



from api.model.ResultVO import *
from schedule import TaskThreadHolder, TaskThread, TaskHandlerContext
from schedule.utils.log import FrameworkHandlerLogger
from client import ScheduleJobTaskLogClient


def stop_job(job_id):
    task_thread = TaskThreadHolder.load_task_thread(job_id)
    if task_thread is not None:
        TaskThreadHolder.remove_task_thread(job_id, "人工手动终止")
        return SUCCESS
    return ResultVO(SUCCESS_CODE, "job thread aleady killed.")


def stop_task(job_id, task_id, is_executing: bool):
    """

    :param job_id:
    :param task_id:
    :return:
    """
    task_thread = TaskThreadHolder.load_task_thread(job_id)

    if task_thread is not None:
        TaskThreadHolder.cancel_task_thread(job_id, task_id, is_executing)
        return SUCCESS
    else:
        ScheduleJobTaskLogClient.update_task_status_to_finished(task_id, ResultVO(FAIL_CODE, "任务不在当前执行队列中,可能执行器被重启.手动强制停止").to_dict(), True)
        return ResultVO(SUCCESS_CODE, "job thread already killed.")


def handle_trigger(task_param, is_sync: bool = False):
    FrameworkHandlerLogger.log_info(f"接收到任务请求,jobId:{task_param.jobId} logId:{task_param.logId}")

    # TODO 这里如果打了断点, 有可能导致生成两个TaskThread
    task_thread: TaskThread = TaskThreadHolder.load_task_thread(task_param.jobId)
    remove_old_reason = None

    if task_thread is not None:
        task_handler = task_thread.get_handler()
    else:
        task_handler = None

    if "BEAN" == task_param.jobType:
        if task_thread is not None and task_handler is not None and task_thread.get_handler() != task_handler:
            remove_old_reason = "更新JobHandler或更换任务模式,终止旧任务线程"
            # 这里不需要显示的终止,在调用 registryTaskThread的时候会自己终止
            task_thread = None
            task_handler = None

        if task_handler is None:
            task_handler = TaskHandlerContext.load_task_handler(task_param.actorHandler)

    elif "JAVA" == task_param.jobType or "SHELL" == task_param.jobType:
        return ResultVO(FAIL_CODE, f"jobType [{task_param.jobType} 此执行器不支持")
    elif "PYTHON" == task_param.jobType:
        # 这里要创建一个 新的 ScriptPython的执行器
        from api.handler.script.ScriptPythonTaskHandler import ScriptPythonTaskHandler
        if task_thread is not None and not (isinstance(task_thread.get_handler(),
                                                       ScriptPythonTaskHandler) and task_thread.get_handler().update_time == task_param.updateTime):
            remove_old_reason = "更新任务逻辑或更换任务模式,终止旧任务线程"
            # 这里不需要显示的终止,在调用 registryTaskThread的时候会自己终止
            task_thread = None
            task_handler = None

        if task_handler is None:
            task_handler = ScriptPythonTaskHandler(task_param.jobId, task_param.updateTime, task_param.scriptJobSource)
    else:
        return ResultVO(FAIL_CODE, f"jobType [{task_param.jobType} 不合法")

    if task_thread is not None:
        # TODO 阻塞策略
        pass

    if task_thread is None:
        task_thread = TaskThreadHolder.registry_task_thread(task_param.jobId, task_handler, remove_old_reason)

    if is_sync:
        # 同步执行
        return task_thread.run_sync(task_param)
    else:
        # 异步执行
        return task_thread.push_trigger_queue(task_param)


def ping():
    return "pong"


def idle(job_id: int):
    task_thread = TaskThreadHolder.load_task_thread(job_id)

    return task_thread is None or not task_thread.is_running_or_has_queue()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/3/6
import threading
from abc import abstractmethod
from datetime import datetime
from unittest.mock import patch, Mock

from munch import Munch

from api.exceptions import IllegalArgumentException
from api.handler.ICacheAbleHandler import ICacheAbleHandler
from api.handler.crawl.AbstractBatchCrawl import AbstractBatchCrawl
from api.handler.crawl.AbstractSimpleCrawl import AbstractSimpleCrawl
from api.handler.crawl.ISimpleScriptCrawl import ISimpleScriptCrawl
from api.handler.outputer.ISimpleOutputer import ISimpleOutputer
from api.handler.script import ScriptHandlerHolder
from api.model.ResultVO import ResultVO, FAIL_CODE, SUCCESS
from client import DataCrawlClient
from schedule.test.suite import BaseInitTestSuite, DataCrawl, PROCESS_LOG_ID, LOG_ID, ScheduleCacheClientMock, \
    TaskLoggerMock
from schedule.utils.log import TaskLogFileAppender
from schedule.utils.log.TaskLogger import TaskLogger


class InnerScriptPythonCrawlTestHandler(ISimpleOutputer, ICacheAbleHandler):

    def __init__(self, simple_script_crawl: ISimpleScriptCrawl):
        super().__init__()
        self._simple_script_crawl: ISimpleScriptCrawl = simple_script_crawl
        self.__code_local_host = threading.local()
        self.__process_id_local = threading.local()
        self.__execution_message_local = threading.local()

    def execute(self, params: dict):
        if params is None or len(params) == 0 or ("site_" not in params) or ("projectCode_" not in params) or (
                "code_" not in params):
            # 不存在crawl的信息,没法执行
            return ResultVO(FAIL_CODE, "执行失败,没有Crawl的信息")

        # 找到对应的dataCrawl
        data_crawl = DataCrawlClient.get(params["site_"], params["projectCode_"], params["code_"])

        if data_crawl is None or data_crawl.type != "CUSTOM":
            return ResultVO(FAIL_CODE, f"执行失败,没有找到Code:{params['code_']}的Crawl,或者该Crawl类型不为Custom")

        # 赋值 日志的上下文
        # log_path = TaskLogFileAppender.prepare_to_log(datetime.now(), LOG_ID)
        # self._simple_script_crawl.task_logger = TaskLogger(log_path)
        # 这里日志输出到控制台，不向日志文件中中记录
        self._simple_script_crawl.task_logger = TaskLoggerMock()

        try:
            self.set_local_code(data_crawl.site + "_" + data_crawl.projectCode + "_" + data_crawl.code)
            self.set_process_id(PROCESS_LOG_ID)
            ScriptHandlerHolder.holder.key = self
            self.__execution_message_local.content = []
            result_list = self._simple_script_crawl.do_crawl(params)
        except Exception as e:
            return ResultVO(FAIL_CODE, f"执行失败,Code:{params['site_']}的Crawl,发生错误:{str(e)}")
        finally:
            # 清理 MessageLocal 和 LocalCode
            self.clear_local_code()
            self.clear_process_id()
            self.__execution_message_local.content = None

            ScriptHandlerHolder.holder.key = None
            self._simple_script_crawl.task_logger = None
        return self._send_results(params, result_list, data_crawl.outputFdCodes)

    def _send_results(self, params, result_list, output_fd_codes):
        """
        发送结果数据
        :param params:
        :param result_list:
        :param output_fd_codes:
        :return:
        """
        if result_list is None or len(result_list) == 0:
            # 搞完了,没的结果,不处理
            return SUCCESS
        # 有结果,就需要从crawl的配置中找到目标的fd,然后调用fd的接口进行保存
        try:
            self.send_output_result(params["code_"], result_list, output_fd_codes)
        except Exception as e:
            self._simple_script_crawl.task_logger(
                f"{params['site_']}_{params['projectCode_']}_{params['code_']} 发送结果数据失败,",e)
            return ResultVO(FAIL_CODE, f"发送结果数据失败,错误:{str(e)}")
        return SUCCESS

    def get_code_thread_local(self):
        return self.__code_local_host

    def get_process_thread_local(self):
        return self.__process_id_local

    def get_execution_message_cache(self):
        return self.__execution_message_local.content


class AbstractSimpleScriptCrawlTestSuite(BaseInitTestSuite):
    """
    crawlSimple脚本测试组件
    """
    # region 需要实现的方法
    @abstractmethod
    def mock_target(self) -> ISimpleScriptCrawl:
        """
        设置crawl实例
        :return:
        """

    @abstractmethod
    def mock_fact_datasource(self) -> dict:
        """
        设置FD的基本定义,不区分数据源和主数据
        :return: dict<code,FactDatasource>
        """

    @abstractmethod
    def mock_crawl(self) -> DataCrawl:
        """
        设置crawl的基本定义
        :return: DataCrawl
        """
    # endregion

    # region 内部逻辑
    def setUp(self):
        target: ISimpleScriptCrawl = self.mock_target()
        if not target:
            raise IllegalArgumentException('测试抓取器实例为空')

        fds = self.mock_fact_datasource()
        if not fds:
            raise IllegalArgumentException('测试数据源为空')

        crawl = self.mock_crawl()
        if not crawl:
            raise IllegalArgumentException('测试抓取器定义为空')

    def __fd_client(self):
        return self._get_datesource_client()

    @patch('schedule.utils.log.TaskLogFileAppender.prepare_to_log')
    @patch('factdatasource.FactDatasourceProxyService.fd_client_proxy')
    @patch('client.DataCrawlClient.get')
    def crawl_execute(self, mock_crawl_get, mock_fd_client, mock_prepare_to_log):
        """
        执行测试
        :return:
        """
        mock_crawl_get.return_value = self.__mock_crawl()
        mock_fd_client.return_value = self.__fd_client()
        mock_prepare_to_log.return_value = None

        crawl = self.mock_crawl()
        crawl_param = {
            'site_': crawl.site,
            'projectCode_': crawl.projectCode,
            'code_': crawl.code
        }
        crawl_param.update(crawl.params)

        target: ISimpleScriptCrawl = self.mock_target()
        crawl_handler = InnerScriptPythonCrawlTestHandler(target)

        if self.distributed_mode():
            # 模拟缓存
            cache_mock = ScheduleCacheClientMock()
            with patch('client.ScheduleCacheClient.set_value', Mock(side_effect=cache_mock.set_value)), \
                 patch('client.ScheduleCacheClient.set_if_absent', Mock(side_effect=cache_mock.set_if_absent)), \
                 patch('client.ScheduleCacheClient.get', Mock(side_effect=cache_mock.get)), \
                 patch('client.ScheduleCacheClient.delete', Mock(side_effect=cache_mock.delete)):
                    self.execute(crawl_handler, crawl_param)
        else:
            self.execute(crawl_handler, crawl_param)

    def __mock_crawl(self):
        crawl = self.mock_crawl()
        return Munch(crawl.to_dict())

    def test_crawl_execute(self):
        """
        测试入口
        :return:
        """
        self.crawl_execute()
    # endregion


class AbstractSimpleCrawlTestSuite(BaseInitTestSuite):
    """
    crawlSimple脚本测试组件
    """

    # region 需要实现的方法
    @abstractmethod
    def mock_target(self) -> AbstractSimpleCrawl:
        """
        设置crawl实例
        :return:
        """

    @abstractmethod
    def mock_fact_datasource(self) -> dict:
        """
        设置FD的基本定义,不区分数据源和主数据
        :return: dict<code,FactDatasource>
        """

    @abstractmethod
    def mock_crawl(self) -> DataCrawl:
        """
        设置crawl的基本定义
        :return: DataCrawl
        """

    def mock_cache_from_process(self) -> dict:
        return {}

    def mock_cache_from_task(self) -> dict:
        return {}
    # endregion

    # region 内部逻辑
    def setUp(self):
        target: AbstractSimpleCrawl = self.mock_target()
        if not target:
            raise IllegalArgumentException('测试抓取器实例为空')

        fds = self.mock_fact_datasource()
        if not fds:
            raise IllegalArgumentException('测试数据源为空')

        crawl = self.mock_crawl()
        if not crawl:
            raise IllegalArgumentException('测试抓取器定义为空')

    def __fd_client(self):
        return self._get_datesource_client()

    def __mock_crawl(self):
        crawl = self.mock_crawl()
        return Munch(crawl.to_dict())

    @patch('schedule.utils.log.TaskLogFileAppender.prepare_to_log')
    @patch('factdatasource.FactDatasourceProxyService.fd_client_proxy')
    @patch('client.DataCrawlClient.get')
    def crawl_execute(self, mock_crawl_get, mock_fd_client, mock_prepare_to_log):
        """
        执行测试
        :return:
        """
        mock_crawl_get.return_value = self.__mock_crawl()
        mock_fd_client.return_value = self.__fd_client()
        mock_prepare_to_log.return_value = None

        crawl = self.mock_crawl()
        crawl_param = {
            'site_': crawl.site,
            'projectCode_': crawl.projectCode,
            'code_': crawl.code,
            '__logId__': LOG_ID,
            '__triggerTime__': datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            '__processLogId__': PROCESS_LOG_ID
        }
        crawl_param.update(crawl.params)

        target: AbstractSimpleCrawl = self.mock_target()

        # 这里日志输出到控制台，不向日志文件中中记录
        target.task_logger = TaskLoggerMock()

        target.set_local_code(f"{crawl.site}_{crawl.projectCode}_{crawl.code}")
        target.set_process_id(PROCESS_LOG_ID)

        if self.distributed_mode():
            # 模拟缓存
            cache_mock = ScheduleCacheClientMock()
            with patch('client.ScheduleCacheClient.set_value', Mock(side_effect=cache_mock.set_value)), \
                 patch('client.ScheduleCacheClient.set_if_absent', Mock(side_effect=cache_mock.set_if_absent)), \
                 patch('client.ScheduleCacheClient.get', Mock(side_effect=cache_mock.get)), \
                 patch('client.ScheduleCacheClient.delete', Mock(side_effect=cache_mock.delete)):

                for key, value in self.mock_cache_from_process().items():
                    target.set_cache_value_from_process(key,value)

                for key, value in self.mock_cache_from_task().items():
                    target.set_cache_value(key,value)

                self.execute(target, crawl_param)
        else:
            self.execute(target, crawl_param)

    def test_crawl_execute(self):
        """
        测试入口
        :return:
        """
        self.crawl_execute()
    # endregion


class AbstractBatchCrawlTestSuite(BaseInitTestSuite):
    """
    crawlSimple脚本测试组件
    """
    # region 需要实现的方法
    @abstractmethod
    def mock_target(self) -> AbstractBatchCrawl:
        """
        设置crawl实例
        :return:
        """

    @abstractmethod
    def mock_fact_datasource(self) -> dict:
        """
        设置FD的基本定义,不区分数据源和主数据
        :return: dict<code,FactDatasource>
        """

    @abstractmethod
    def mock_crawl(self) -> DataCrawl:
        """
        设置crawl的基本定义
        :return: DataCrawl
        """

    def mock_cache_from_process(self) -> dict:
        return {}

    def mock_cache_from_task(self) -> dict:
        return {}
    # endregion

    # region 内部逻辑
    def setUp(self):
        target: AbstractBatchCrawl = self.mock_target()
        if not target:
            raise IllegalArgumentException('测试抓取器实例为空')

        fds = self.mock_fact_datasource()
        if not fds:
            raise IllegalArgumentException('测试数据源为空')

        crawl = self.mock_crawl()
        if not crawl:
            raise IllegalArgumentException('测试抓取器定义为空')

    def __fd_client(self):
        return self._get_datesource_client()

    def __mock_crawl(self):
        crawl = self.mock_crawl()
        return Munch(crawl.to_dict())

    @patch('schedule.utils.log.TaskLogFileAppender.prepare_to_log')
    @patch('factdatasource.FactDatasourceProxyService.fd_client_proxy')
    @patch('client.DataCrawlClient.get')
    def crawl_execute(self, mock_crawl_get, mock_fd_client, mock_prepare_to_log):
        """
        执行测试
        :return:
        """
        mock_crawl_get.return_value = self.__mock_crawl()
        mock_fd_client.return_value = self.__fd_client()
        mock_prepare_to_log.return_value = None

        crawl = self.mock_crawl()
        crawl_param = {
            'site_': crawl.site,
            'projectCode_': crawl.projectCode,
            'code_': crawl.code,
            '__logId__': LOG_ID,
            '__triggerTime__': datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            '__processLogId__': PROCESS_LOG_ID
        }
        crawl_param.update(crawl.params)

        target: AbstractBatchCrawl = self.mock_target()

        # 这里日志输出到控制台，不向日志文件中中记录
        target.task_logger = TaskLoggerMock()

        target.set_local_code(f"{crawl.site}_{crawl.projectCode}_{crawl.code}")
        target.set_process_id(PROCESS_LOG_ID)

        if self.distributed_mode():
            # 模拟缓存
            cache_mock = ScheduleCacheClientMock()
            with patch('client.ScheduleCacheClient.set_value', Mock(side_effect=cache_mock.set_value)), \
                 patch('client.ScheduleCacheClient.set_if_absent', Mock(side_effect=cache_mock.set_if_absent)), \
                 patch('client.ScheduleCacheClient.get', Mock(side_effect=cache_mock.get)), \
                 patch('client.ScheduleCacheClient.delete', Mock(side_effect=cache_mock.delete)):

                for key, value in self.mock_cache_from_process().items():
                    target.set_cache_value_from_process(key,value)

                for key, value in self.mock_cache_from_task().items():
                    target.set_cache_value(key,value)

                self.execute(target, crawl_param)
        else:
            self.execute(target, crawl_param)

    def test_crawl_execute(self):
        """
        测试入口
        :return:
        """
        self.crawl_execute()
    # endregion



"""
schedule相关的操作
"""
import requests
from munch import Munch

from api.exceptions import ServiceInnerException, AuthException
from client import check_instance_avaliable, do_service


def add_job(job_info):
    """
    新增计划
    :param job_info: jobInfo对象
    :return: 如果成功返回True,否则返回False
    """
    if job_info is None:
        return False

    check_instance_avaliable()

    request = None
    if not isinstance(job_info, Munch):
        # 说明不是 Munch类型的 判断是否是Dict的
        if isinstance(job_info, dict):
            # 如果是dict类型的,直接发
            request = job_info
    else:
        # 说明是munch的,那么就转成Dict的
        request = job_info.toDict()

    try:
        r = do_service(f"remote/js-service/", data=request, method="post")
        return r
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def reschedule(job_info):
    """
    更新计划
    :param job_info:
    :return:
    """
    if job_info is None:
        return False

    check_instance_avaliable()

    request = None
    if not isinstance(job_info, Munch):
        # 说明不是 Munch类型的 判断是否是Dict的
        if isinstance(job_info, dict):
            # 如果是dict类型的,直接发
            request = job_info
    else:
        # 说明是munch的,那么就转成Dict的
        request = job_info.toDict()

    try:
        r = do_service(f"remote/js-service/", data=request, method="put")
        return r
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code < 500:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def remove(job_id):
    """
    删除一个计划
    :param job_id:
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/js-service/{job_id}",method="delete")
        if  r is not None:
            return r
        else:
            return False
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def pause(job_id):
    """
    暂停计划
    :param job_id:
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/js-service/{job_id}/pause", method="post")
        if  r is not None:
            return r
        else:
            return False
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def resume(job_id):
    """
    恢复计划
    :param job_id:
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/js-service/{job_id}/resume", method="post")
        if r is not None:
            return r
        else:
            return False
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def stop(job_id):
    """
    停止计划
    :param job_id:
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/js-service/{job_id}/stop", method="post")
        if r is not None:
            return r
        else:
            return False
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def trigger_job(job_id, params=None):
    """
    触发一个计划
    :param job_id:
    :param params:
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/js-service/{job_id}/trigger",data=params, method="post")
        if r is not None:
            return r
        else:
            return False
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def list_jobs():
    """
    列出所有的计划
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/js-service")
        if r is not None:
            if isinstance(r, list):
                return [Munch(x) for x in r]
            else:
                return Munch(r)
        else:
            return None
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def get_jobs(actor_name):
    """
    找到某一个类型的所有计划
    :param actor_name:
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/js-service/{actor_name}")
        if r is not None:
            if isinstance(r, list):
                return [Munch(x) for x in r]
            else:
                return Munch(r)
        else:
            return None
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e


def find_by_site_and_code(site, project_code, code):
    """
    返回某一个计划
    :param site:
    :param project_code:
    :param code:
    :return:
    """
    check_instance_avaliable()

    try:
        r = do_service(f"remote/js-service/{site}/{project_code}/{code}")
        if r is not None:
            return Munch(r)
        else:
            return None
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 500:
            # 说明服务器端报错了
            raise ServiceInnerException(e.response._content.decode('utf-8'))
        elif e.response.status_code >= 400 and e.response.status_code<500 and e.response.status_code!=404:
            # 说明是认证相关的错误
            raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
        raise e



from api.annotation import annotation


def registry_task_handler(name: str, job_handler):
    annotation.TASK_HANDLERS[name] = job_handler
    return job_handler


def load_task_handler(name: str):
    return annotation.TASK_HANDLERS.get(name)



"""
yml文件注册功能
"""
import os

import yaml

REGISTRY_PROPERTIES = dict()


def load_registry_properties(file, yaml_name: str):
    yaml = read_yaml_file(file, yaml_name)
    if yaml is None:
        return
    global REGISTRY_PROPERTIES
    REGISTRY_PROPERTIES = yaml["sobeyficus"]["registry"]


def read_yaml_file(file, yaml_name: str, profile: str = 'default'):
    """
    读取文件路径中的yaml文件
    :param yaml_name:
    :return:
    """
    # 获取当前文件路径
    filePath = os.path.dirname(file)

    # 这里采用一层一层的递归上去找是否有yaml_name的文件
    # 获取配置文件的路径
    yamlPath = _find_path(filePath, yaml_name)
    if yamlPath is None:
        # 如果文件不存在,忽略
        return None

    # 加上 ,encoding='utf-8'，处理配置文件中含中文出现乱码的情况。
    with open(yamlPath, 'r', encoding='utf-8') as f:
        cont = f.read()

    # 加载多个对象
    docs = yaml.load_all(cont, Loader=yaml.FullLoader)

    # 如果数量大于1,并且填了profile的.那么就按照profile的做合并. key是 spring.profiles
    result = None
    for doc in docs:
        if result is None:
            result = doc
            continue
        if ("spring" in doc and "profiles" in doc["spring"]) and (
                (isinstance(doc["spring"]["profiles"], list) and profile in doc["spring"]["profiles"]) or (
                profile == doc["spring"]["profiles"])):
            # 找到是相同的,进行合并
            from config import annotation
            annotation.deep_search_and_merge(result, doc)

    # 返回yaml文件对象
    return result


def _find_path(file_path, yaml_name):
    """
    递归查找文件
    :param file_path:
    :param yaml_name:
    :return:
    """
    # 这里window中取出来的路径也可能是 / 分割的，这里简单重写下
    r_path = os.path.join(file_path, yaml_name)
    if os.path.exists(r_path):
        return r_path
    else:
        folder_path = os.path.split(file_path)[0]
        if folder_path == file_path:
            return None
        else:
            return _find_path(folder_path, yaml_name)


def _get_separator():
    """
    判断是windows还是其他的,系统不一样 separtor不一样
    :return:
    """
    import platform
    if 'Windows' in platform.system():
        separator = '\\'
    else:
        separator = '/'
    return separator



import json

import yaml
from munch import Munch

from api.exceptions import IllegalArgumentException
from rule_engine import *

# region 私有方法
from rule_engine import RuleEngineCustomScriptFactory


def __find_class_from_str(clazz:str):
    """
    从str获取类class
    :param clazz:
    :return:
    """
    ss = clazz.rsplit('.', 1)
    try:
        if len(ss) == 2:
            package = __import__(ss[0])
            return getattr(package, ss[1])
        else:
            package = __import__("")
            return getattr(package, clazz)
    except:
        # 有可能类加载不出来,因为是JAVA的,或者类路径写错了
        return None

def __parseStr(rule_groups:str,func):
    """
    从字符串转换成为rule_group_info对象
    :param rule_groups:
    :param func: 解析字符串到dict的函数
    :return:
    """
    dic = func(rule_groups)

    yml = Munch.fromDict(dic)

    results = []

    try:
        cof = yml.sobeyficus.ruleengine
        if len(cof.groups) == 0:
            return results
    except AttributeError:
        # 说明没有配置这个键,直接返回
        raise IllegalArgumentException("解析规则组内容失败,没有节点:sobeyficus.ruleengine ")

    # 开始构造模型
    groups = yml.sobeyficus.ruleengine.groups

    # 在配置文件中找到了相关的规则配置,那么就需要加工,并放入到规则引擎中
    for group in groups:
        rule_group_info = __build_rule_group(group)
        results.append(rule_group_info)

    return results

# endregion


# region 构建模型
def __build_rule_then(then):
    thens = []

    if then is not None and len(then) > 0:
        for th in then:
            t = RuleAction()
            thens.append(t)

            try:
                t.action = th.action
            except AttributeError:
                pass

            try:
                t.type = RuleActionTypeEnum[th.type]
            except AttributeError:
                pass

            try:
                t.params = th.params
            except AttributeError:
                pass

    return thens


def __build_rule_when(when):
    whens = []

    if when is not None and len(when) > 0:
        for wh in when:
            w = RuleConditionInfo()
            whens.append(w)

            try:
                w.source = wh.source
            except AttributeError:
                pass

            try:
                w.target = wh.target
            except AttributeError:
                pass

            try:
                w.operation = RuleConditionOperationEnum[wh.operation]
            except AttributeError:
                pass

            try:
                w.enable = wh.enable
            except AttributeError:
                pass

            try:
                w.relation = RuleConditionRelationEnum[wh.relation]
            except AttributeError:
                pass

            try:
                w.subs = __build_rule_when(wh.subs)
            except AttributeError:
                pass

    return whens


def __build_rule_info(rules):
    """
    构建ruleInfo对象
    :param rules:
    :return:
    """
    rs = []

    if rules is not None and len(rules)>0:
        for rule in rules:
            r = RuleInfo()
            rs.append(r)

            try:
                r.name = rule.name
            except AttributeError:
                pass

            try:
                r.description = rule.description
            except AttributeError:
                pass

            try:
                r.enable = rule.enable
            except AttributeError:
                pass

            try:
                r.then = __build_rule_then(rule.then)
            except AttributeError:
                pass

            try:
                r.when = __build_rule_when(rule.when)
            except AttributeError:
                pass

    return rs

def __build_rule_group(group:Munch):
    """
    构建 rulegroup对象
    :param group:
    :return:
    """
    rule_group_info = RuleGroupInfo()
    try:
        rule_group_info.name = group.name
    except AttributeError:
        pass

    try:
        rule_group_info.project = group.project
    except AttributeError:
        pass

    try:
        rule_group_info.match_all = group.matchAll
    except AttributeError:
        pass

    try:
        rule_group_info.enable = group.enable
    except AttributeError:
        pass

    try:
        rule_group_info.default_rule = group.defaultRule
    except AttributeError:
        pass

    try:
        rule_group_info.rules = __build_rule_info(group.rules)
    except AttributeError:
        pass

    return rule_group_info
# endregion


# region 规则校验
def __inner_check_rule_group(rule_group_info):
    """
    校验规则组
    :param rule_group_info:
    :return:
    """
    if rule_group_info.name is None or rule_group_info.name=="":
        raise IllegalArgumentException("校验规则组失败:名称为空")

    if rule_group_info.project is None or rule_group_info.project=="":
        raise IllegalArgumentException("校验规则组失败:项目为空")

    if rule_group_info.rules is None or len(rule_group_info.rules)==0:
        raise IllegalArgumentException("校验规则组失败:规则为空")

    if rule_group_info.default_rule is not None:
        for rule in rule_group_info.rules:
            if rule.name == rule_group_info.default_rule:
                return
        raise IllegalArgumentException(f"校验规则组失败:默认规则{rule_group_info.default_rule} 不存在")


def __inner_check_rule(rule:RuleInfo):
    """
    校验规则
    :param rule:
    :return:
    """
    if rule.name is None or rule.name=="":
        raise IllegalArgumentException("校验规则失败:名称为空")

    if rule.when is None or len(rule.when)==0:
        raise IllegalArgumentException(f"校验规则{rule.name}失败:规则条件为空")


def __inner_check_rule_condition(when:RuleConditionInfo):
    """
    校验条件
    :param when:
    :return:
    """
    if len(when.subs)==0 and when.operation is None:
        raise IllegalArgumentException(f"校验规则条件失败:条件比对方式为空")

    if len(when.subs)>0 and when.operation is not None:
        raise IllegalArgumentException(f"校验规则条件失败:有子条件,条件比对方式必须为空")

    if when.operation is not None:
        if not (when.operation in [RuleConditionOperationEnum.MVEL,RuleConditionOperationEnum.AVIATOR,RuleConditionOperationEnum.BEAN,RuleConditionOperationEnum.CLASS,RuleConditionOperationEnum.PY_BEAN,RuleConditionOperationEnum.PY_CLASS,RuleConditionOperationEnum.PY_EXP] or (when.source is not None and when.source !="")):
            raise IllegalArgumentException(f"校验规则条件失败:规则条件为空或不合法")

    if when.operation is not None:
        # 校验每一个方式的合法性
        if when.operation == RuleConditionOperationEnum.PY_BEAN:
            # 校验源码加载
            RuleEngineCustomScriptFactory.load_instance(when.target,"IRuleConditionOperation")
        elif when.operation == RuleConditionOperationEnum.PY_CLASS:
            # 校验获取类路径
            clazz = __find_class_from_str(when.target)
            if clazz is None or not issubclass(clazz,IRuleConditionOperation):
                raise IllegalArgumentException(f"校验规则条件失败:规则不合法")
        # 其他类型的可以不校验

    if len(when.subs)>0:
        for sub in when.subs:
            # 递归校验子条件
            __inner_check_rule_condition(sub)

def __inner_check_rule_action(rule_action:RuleAction):
    """
    校验结果
    :param then:
    :return:
    """
    if rule_action.type is None:
        raise IllegalArgumentException("校验规则条件失败:执行类型为空")

    if rule_action.type is not None:
        if not ((rule_action.type in [RuleActionTypeEnum.MVEL,RuleActionTypeEnum.CLASS,RuleActionTypeEnum.BEAN,RuleActionTypeEnum.AVIATOR,RuleActionTypeEnum.PY_CLASS,RuleActionTypeEnum.PY_BEAN,RuleActionTypeEnum.PY_EXP] and (rule_action.action is not None and rule_action.action!="")) or rule_action.type==RuleActionTypeEnum.DEFAULT):
            raise IllegalArgumentException("校验规则条件失败:执行动作为空")

    if rule_action.type is not None:
        # 校验每一个方式的合法性
        if rule_action.type == RuleActionTypeEnum.PY_BEAN:
            # 校验源码加载
            RuleEngineCustomScriptFactory.load_instance(rule_action.action,"IRuleAction")
        elif rule_action.type == RuleActionTypeEnum.PY_CLASS:
            # 校验获取类路径
            clazz = __find_class_from_str(rule_action.action)
            if clazz is None or not issubclass(clazz, IRuleAction):
                raise IllegalArgumentException(f"校验规则条件失败:规则不合法")
        # 其他类型的可以不校验

def __check_rule_group(rule_group_info:RuleGroupInfo):
    """
    校验rule_group的合法性
    :param rule_group_info:
    :return:
    """
    # 校验规则组
    __inner_check_rule_group(rule_group_info)

    for rule in rule_group_info.rules:
        __inner_check_rule(rule)

        for when in rule.when:
            __inner_check_rule_condition(when)

        for then in rule.then:
            __inner_check_rule_action(then)
# endregion


# region 规则补全
def __inner_rebuild_rule_condition(when:RuleConditionInfo):
    if when.operation == None:
        when.operation = RuleConditionOperationEnum.DO_NOTHING

    # 主要是处理IRuleConditionOperation 和 targetProvider
    when.rule_operation = when.operation.condition_operation()

    if when.operation == RuleConditionOperationEnum.PY_BEAN or when.operation == RuleConditionOperationEnum.PY_CLASS:
        if when.rule_operation is not None:
            raise IllegalArgumentException("operation已经指定,不能重复自定义")

        if when.operation == RuleConditionOperationEnum.PY_BEAN:
            when.rule_operation = RuleEngineCustomScriptFactory.load_instance(when.target,"IRuleConditionOperation")
        elif when.operation == RuleConditionOperationEnum.PY_CLASS:
            clazz = __find_class_from_str(when.target)
            when.rule_operation = clazz()

    # 递归构造子条件
    if len(when.subs)>0:
        for sub in when.subs:
            __inner_rebuild_rule_condition(sub)

def __inner_rebuild_rule_action(then:RuleAction):
    then.rule_action = then.type.actions()

    if then.type == RuleActionTypeEnum.PY_BEAN or then.type == RuleActionTypeEnum.PY_CLASS:
        if then.rule_action is not None:
            raise IllegalArgumentException("action已经指定,不能重复自定义")

        if then.type == RuleActionTypeEnum.PY_BEAN:
            then.rule_action = RuleEngineCustomScriptFactory.load_instance(then.action,"IRuleAction")
        elif then.type == RuleActionTypeEnum.PY_CLASS:
            clazz = __find_class_from_str(then.action)
            then.rule_action = clazz()


def __rebuild_rule_group(rule_group_info:RuleGroupInfo):

    for rule in rule_group_info.rules:

        for when in rule.when:
            __inner_rebuild_rule_condition(when)

        for then in rule.then:
            __inner_rebuild_rule_action(then)
# endregion


def __initialize(group)->RuleGroupInfo:

    rule_group_info = __build_rule_group(group)

    __check_rule_group(rule_group_info)

    __rebuild_rule_group(rule_group_info)

    return rule_group_info


def initialize_from_obj(rule_group_info:RuleGroupInfo)->RuleGroupInfo:
    """
    从构造器构造的RuleGroupInfo实例,初始化规则引擎
    :param rule_group_info:
    :return:
    """
    if rule_group_info is None:
        raise IllegalArgumentException("初始化规则引擎失败,传入的规则组实例为空")

    # 检测规则是否合法
    __check_rule_group(rule_group_info)

    # 构造info中的必填项
    __rebuild_rule_group(rule_group_info)

    from rule_engine import RuleEngine
    RuleEngine.register_rule_group(rule_group_info)

    return rule_group_info

def initialize_from_str(rule_groups:str):
    """
    从规则字符串中,初始化规则引擎
    :param rule_groups:  可以是 json 也可以是 yaml
    :return:
    """
    if rule_groups is None or rule_groups.strip()=="":
        raise IllegalArgumentException("初始化规则引擎失败,传入的规则组字符串为空")

    rule_groups = rule_groups.strip()

    if rule_groups.startswith("{") and rule_groups.endswith("}"):
        # 说明可能是json格式的
        result = __parseStr(rule_groups,lambda x:json.loads(x))
    else:
        # 说明是yaml格式的
        result = __parseStr(rule_groups, lambda x: yaml.load(x))

    if result is not None:
        for ruleGroupInfo in result:
            initialize_from_obj(ruleGroupInfo)

    return result


def initialize_from_remote():
    """
    从yaml配置中,初始化规则引擎
    :return:
    """
    from config import annotation

    # 获取远程配置中的配置项.判断是否有规则引擎相关的配置
    yml = Munch.fromDict(annotation.REMOTE_YML_CONFIG)

    try:
        cof = yml.sobeyficus.ruleengine
        if len(cof.groups) == 0 :
            return
    except AttributeError:
        # 说明没有配置这个键,直接返回
        return

    # 开始构造模型
    groups = yml.sobeyficus.ruleengine.groups

    # 在配置文件中找到了相关的规则配置,那么就需要加工,并放入到规则引擎中
    from rule_engine import RuleEngine
    for group in groups:
        rule_group_info = __initialize(group)
        RuleEngine.register_rule_group(rule_group_info)

